"""Astronomy reference frames."""
# ruff:noqa: N806

__all__: list[str] = []


from typing import Literal, TypeAlias

from jaxtyping import Array, Shaped
from plum import convert, dispatch

import quaxed.numpy as jnp
import unxt as u

from .galactocentric import Galactocentric
from .icrs import ICRS
from coordinax._src.angle import Angle
from coordinax._src.distance import Distance
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.operators.identity import IdentityOperator
from coordinax._src.operators.sequential import OperatorSequence
from coordinax._src.vectors.base import AbstractVel
from coordinax._src.vectors.d3 import AbstractPos3D, CartesianPos3D, CartesianVel3D

ScalarAngle: TypeAlias = Shaped[u.Quantity["angle"] | Angle, ""]
RotationMatrix: TypeAlias = Shaped[Array, "3 3"]
LengthVector: TypeAlias = Shaped[u.Quantity["length"], "3"] | Shaped[Distance, "3"]
VelocityVector: TypeAlias = Shaped[u.Quantity["speed"], "3"]


@dispatch
def frame_transform_op(from_frame: ICRS, to_frame: ICRS, /) -> IdentityOperator:
    """Return an identity operator for the ICRS->ICRS transformation.

    Examples
    --------
    >>> import coordinax.frames as cxf
    >>> icrs_frame = cxf.ICRS()
    >>> frame_op = cxf.frame_transform_op(icrs_frame, icrs_frame)
    >>> frame_op
    IdentityOperator()

    """
    return IdentityOperator()


# ---------------------------------------------------------------


@dispatch
def frame_transform_op(
    from_frame: Galactocentric, to_frame: Galactocentric, /
) -> OperatorSequence:
    """Return a sequence of operators for the Galactocentric frame self transformation.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.frames as cxf

    >>> gcf_frame = cxf.Galactocentric()
    >>> frame_op = cxf.frame_transform_op(gcf_frame, gcf_frame)
    >>> frame_op
    OperatorSequence(operators=(IdentityOperator(),))

    >>> gcf_frame2 = cxf.Galactocentric(roll=u.Quantity(10, "deg"))
    >>> frame_op2 = cxf.frame_transform_op(gcf_frame, gcf_frame2)
    >>> frame_op2
    OperatorSequence(
      operators=( _GCF2ICRSOperator( ... ), _ICRS2GCFOperator( ... ) )
    )

    """
    if from_frame == to_frame:
        return OperatorSequence((IdentityOperator(),))

    # TODO: not go through ICRS for the self-transformation
    return _GCF2ICRSOperator(from_frame) | _ICRS2GCFOperator(to_frame)


# ---------------------------------------------------------------


def passive_rotation_matrix_x(angle: ScalarAngle) -> RotationMatrix:
    """Passive rotation matrix about the x-axis."""
    c = jnp.cos(u.ustrip("rad", angle))
    s = jnp.sin(u.ustrip("rad", angle))
    return jnp.asarray([[1.0, 0, 0], [0, c, s], [0, -s, c]])


def passive_rotation_matrix_y(angle: ScalarAngle) -> RotationMatrix:
    """Passive rotation matrix about the y-axis."""
    c = jnp.cos(u.ustrip("rad", angle))
    s = jnp.sin(u.ustrip("rad", angle))
    return jnp.asarray([[c, 0, -s], [0, 1.0, 0], [s, 0, c]])


def passive_rotation_matrix_z(angle: ScalarAngle) -> RotationMatrix:
    """Passive rotation matrix about the z-axis."""
    c = jnp.cos(u.ustrip("rad", angle))
    s = jnp.sin(u.ustrip("rad", angle))
    return jnp.asarray([[c, s, 0], [-s, c, 0], [0, 0, 1.0]])


def _icrs_cartesian_to_gcf_cartesian_matrix_vectors(
    frame: Galactocentric, /
) -> tuple[RotationMatrix, LengthVector, VelocityVector]:
    """ICRS->GCF transformation matrices and offsets."""
    # rotation matrix to align x(ICRS) with the vector to the Galactic center
    mat1 = passive_rotation_matrix_y(-frame.galcen.lat)
    mat2 = passive_rotation_matrix_z(frame.galcen.lon)
    # extra roll away from the Galactic x-z plane
    mat0 = passive_rotation_matrix_x(frame.roll0 - frame.roll)

    # construct transformation matrix and use it
    R = mat0 @ mat1 @ mat2

    # Now need to translate by Sun-Galactic center distance around x' and
    # rotate about y' to account for tilt due to Sun's height above the plane
    z_d = u.ustrip("", frame.z_sun / frame.galcen.distance)  # [radian]
    H = passive_rotation_matrix_y(-u.Quantity(jnp.asin(z_d), "rad"))

    # compute total matrices
    A = H @ R
    offset = -H @ (frame.galcen.distance * jnp.asarray([1.0, 0.0, 0.0]))

    return A, offset, convert(frame.galcen_v_sun, u.Quantity)


class _ICRS2GCFOperator(AbstractOperator):
    """Transform from ICRS to Galactocentric frame.

    .. warning::

        This operator is temporary and will be replaced as an operator Sequence
        of the individual transformations.

    Examples
    --------
    For this example we compare against the Astropy implementation of the
    frame transformation:

    >>> import unxt as u
    >>> import coordinax as cx
    >>> import astropy.coordinates as apyc

    >>> vega = apyc.SkyCoord(
    ...     ra=279.23473479 * u.unit("deg"), dec=38.78368896 * u.unit("deg"),
    ...     distance=25 * u.unit("pc"),
    ...     pm_ra_cosdec=200 * u.unit("mas / yr"), pm_dec=-286 * u.unit("mas / yr"),
    ...     radial_velocity=-13.9 * u.unit("km / s"))
    >>> print(vega)
    <SkyCoord (ICRS): (ra, dec, distance) in (deg, deg, pc)
        (279.23473479, 38.78368896, 25.)
     (pm_ra_cosdec, pm_dec, radial_velocity) in (mas / yr, mas / yr, km / s)
        (200., -286., -13.9)>

    >>> apy_gcf = apyc.Galactocentric()
    >>> apy_gcf
    <Galactocentric Frame (galcen_coord=<ICRS Coordinate: (ra, dec) in deg
        (266.4051, -28.936175)>, galcen_distance=8.122 kpc, galcen_v_sun=(12.9, 245.6, 7.78) km / s, z_sun=20.8 pc, roll=0.0 deg)>

    >>> vega.transform_to(apy_gcf)
    <SkyCoord (Galactocentric: ...): (x, y, z) in pc
        (-8112.89970167, 21.79911216, 29.01384942)
     (v_x, v_y, v_z) in km / s
        (34.06711868, 234.61647066, -28.75976702)>

    Now we can use the `_ICRS2GCFOperator` to perform the same transformation:

    >>> vega_q = cx.LonLatSphericalPos.from_(vega.icrs.data)
    >>> vega_p = cx.LonCosLatSphericalVel.from_(vega.icrs.data.differentials["s"])

    >>> icrs_frame = cx.frames.ICRS()
    >>> gcf_frame = cx.frames.Galactocentric.from_(apy_gcf)

    >>> frame_op = cx.frames.frame_transform_op(icrs_frame, gcf_frame)
    >>> vega_gcf_q, vega_gcf_p = frame_op(vega_q, vega_p)
    >>> print(vega_gcf_q)
    <CartesianPos3D (x[pc], y[pc], z[pc])
        [-8112.899    21.799    29.014]>
    >>> print(vega_gcf_p.uconvert({"speed": "km/s"}))
    <CartesianVel3D (d_x[km / s], d_y[km / s], d_z[km / s])
        [ 34.067 234.616 -28.76 ]>

    It matches!

    """  # noqa: E501

    gcf: Galactocentric

    @property
    def is_inertial(self) -> Literal[True]:
        """Return that the operator is inertial.

        Examples
        --------
        >>> import coordinax.frames as cxf
        >>> frame_op = cxf.frame_transform_op(cxf.ICRS(), cxf.Galactocentric())
        >>> frame_op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> AbstractOperator:
        """Return the inverse operator -- Galactocentric to ICRS.

        Examples
        --------
        >>> import coordinax.frames as cxf

        >>> icrs, gcf = cxf.ICRS(), cxf.Galactocentric()
        >>> frame_op = cxf.frame_transform_op(icrs, gcf)

        >>> frame_op.inverse == cxf.frame_transform_op(gcf, icrs)
        Array(True, dtype=bool)

        """
        return _GCF2ICRSOperator(self.gcf)

    def _call_q(self, q: LengthVector, /) -> LengthVector:
        A, offset, _ = _icrs_cartesian_to_gcf_cartesian_matrix_vectors(self.gcf)

        return A @ q + offset

    @dispatch
    def __call__(self, q: LengthVector, /) -> LengthVector:
        return self._call_q(q)

    @dispatch
    def __call__(
        self, q: LengthVector, p: VelocityVector, /
    ) -> tuple[LengthVector, VelocityVector]:
        """Transform q and p from ICRS Cartesian -> GCF Cartesian."""
        # Compute the transformation matrix and offsets
        A, offset, offset_v = _icrs_cartesian_to_gcf_cartesian_matrix_vectors(self.gcf)

        # Apply the transformation to the position
        qp = A @ q + offset

        # Apply the transformation to the velocity
        jac = u.experimental.jacfwd(self._call_q, argnums=0, units=(q.unit,))(q)
        pp = jac @ p + offset_v

        return qp, pp

    @dispatch
    def __call__(
        self, qvec: AbstractPos3D, pvec: AbstractVel
    ) -> tuple[AbstractPos3D, AbstractVel]:
        p = convert(pvec.represent_as(CartesianVel3D, qvec), u.Quantity)

        qp, pp = self(convert(qvec, u.Quantity), p)
        qpvec = CartesianPos3D.from_(qp)
        ppvec = CartesianVel3D.from_(pp)

        return qpvec, ppvec


@dispatch  # type: ignore[misc]
def frame_transform_op(
    from_frame: ICRS, to_frame: Galactocentric, /
) -> _ICRS2GCFOperator:
    """Return an ICRS to Galactocentric frame transformation operator.

    Examples
    --------
    >>> import coordinax.frames as cxf
    >>> icrs_frame = cxf.ICRS()
    >>> gcf_frame = cxf.Galactocentric()
    >>> frame_op = cxf.frame_transform_op(icrs_frame, gcf_frame)
    >>> frame_op
    _ICRS2GCFOperator(
      gcf=Galactocentric(
        galcen=LonLatSphericalPos( ... ),
        roll=Quantity[...](value=weak_i32[], unit=Unit("deg")),
        z_sun=Quantity[...](value=weak_f32[], unit=Unit("pc")),
        galcen_v_sun=CartesianVel3D( ... )
      )
    )

    """
    return _ICRS2GCFOperator(to_frame)


# ---------------------------------------------------------------


def _gcf_cartesian_to_icrs_cartesian_matrix_vectors(
    frame: Galactocentric, /
) -> tuple[RotationMatrix, LengthVector, VelocityVector]:
    """GCF->ICRS transformation matrices and offsets."""
    # ICRS -> GCF
    A, offset, offset_v = _icrs_cartesian_to_gcf_cartesian_matrix_vectors(frame)

    # GCF -> ICRS
    A = A.T  # (A^-1 = A^T)
    offset = A @ (-offset)
    offset_v = A @ (-offset_v)

    return A, offset, offset_v


class _GCF2ICRSOperator(AbstractOperator):
    """Transform from Galactocentric to ICRS frame.

    .. warning::

        This operator is temporary and will be replaced as an operator Sequence
        of the individual transformations.

    Examples
    --------
    For this example we compare against the Astropy implementation of the
    frame transformation:

    >>> import unxt as u
    >>> import coordinax as cx
    >>> import astropy.coordinates as apyc

    >>> apy_gcf = apyc.Galactocentric()
    >>> apy_gcf
    <Galactocentric Frame (galcen_coord=<ICRS Coordinate: (ra, dec) in deg
        (266.4051, -28.936175)>, galcen_distance=8.122 kpc, galcen_v_sun=(12.9, 245.6, 7.78) km / s, z_sun=20.8 pc, roll=0.0 deg)>

    >>> vega = apyc.SkyCoord(
    ...     ra=279.23473479 * u.unit("deg"), dec=38.78368896 * u.unit("deg"),
    ...     distance=25 * u.unit("pc"),
    ...     pm_ra_cosdec=200 * u.unit("mas / yr"), pm_dec=-286 * u.unit("mas / yr"),
    ...     radial_velocity=-13.9 * u.unit("km / s")
    ... ).transform_to(apy_gcf)
    >>> print(vega)
    <SkyCoord (Galactocentric: ...): (x, y, z) in pc
        (-8112.89970167, 21.79911216, 29.01384942)
     (v_x, v_y, v_z) in km / s
        (34.06711868, 234.61647066, -28.75976702)>

    >>> vega.transform_to(apyc.ICRS())
    <SkyCoord (ICRS): (ra, dec, distance) in (deg, deg, pc)
        (279.23473479, 38.78368896, 25.)
     (pm_ra_cosdec, pm_dec, radial_velocity) in (mas / yr, mas / yr, km / s)
        (200., -286., -13.9)>

    Now we can use the `_GCF2ICRSOperator` to perform the same transformation:

    >>> vega_q = cx.CartesianPos3D.from_(vega.galactocentric.data)
    >>> vega_p = cx.CartesianVel3D.from_(vega.galactocentric.data.differentials["s"])

    >>> icrs_frame = cx.frames.ICRS()
    >>> gcf_frame = cx.frames.Galactocentric.from_(apy_gcf)

    >>> frame_op = cx.frames.frame_transform_op(gcf_frame, icrs_frame)
    >>> vega_icrs_q, vega_icrs_p = frame_op(vega_q, vega_p)

    >>> vega_icrs_q = vega_icrs_q.represent_as(cx.LonLatSphericalPos)
    >>> print(vega_icrs_q.uconvert({"angle": "deg", "length": "pc"}))
    <LonLatSphericalPos (lon[deg], lat[deg], distance[pc])
        [279.235  38.785  25.   ]>

    >>> vega_icrs_p = vega_icrs_p.represent_as(cx.LonCosLatSphericalVel, vega_icrs_q)
    >>> print(vega_icrs_p.uconvert({"angular speed": "mas / yr", "speed": "km/s"}))
    <LonCosLatSphericalVel (d_lon_coslat[mas / yr], d_lat[mas / yr], d_distance[km / s])
        [ 200.002 -286.002  -13.9  ]>

    It matches!

    """  # noqa: E501

    gcf: Galactocentric

    @property
    def is_inertial(self) -> Literal[True]:
        """Return that the operator is inertial.

        Examples
        --------
        >>> import coordinax.frames as cxf
        >>> frame_op = cxf.frame_transform_op(cxf.Galactocentric(), cxf.ICRS())
        >>> frame_op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> AbstractOperator:
        """Return the inverse operator -- ICRS to Galactocentric.

        Examples
        --------
        >>> import coordinax.frames as cxf

        >>> icrs, gcf = cxf.ICRS(), cxf.Galactocentric()
        >>> frame_op = cxf.frame_transform_op(gcf, icrs)

        >>> frame_op.inverse == cxf.frame_transform_op(icrs, gcf)
        Array(True, dtype=bool)

        """
        return _ICRS2GCFOperator(self.gcf)

    def _call_q(self, q: LengthVector, /) -> LengthVector:
        A, offset, _ = _gcf_cartesian_to_icrs_cartesian_matrix_vectors(self.gcf)

        return A @ q + offset

    @dispatch
    def __call__(self, q: LengthVector, /) -> LengthVector:
        return self._call_q(q)

    @dispatch
    def __call__(
        self, q: LengthVector, p: VelocityVector, /
    ) -> tuple[LengthVector, VelocityVector]:
        """Transform q and p from GCF Cartesian -> ICRS Cartesian."""
        # Compute the transformation matrix and offsets
        A, offset, offset_v = _gcf_cartesian_to_icrs_cartesian_matrix_vectors(self.gcf)

        # Apply the transformation to the position
        qp = A @ q + offset

        # Apply the transformation to the velocity
        jac = u.experimental.jacfwd(self._call_q, argnums=0, units=(q.unit,))(q)
        pp = jac @ p + offset_v

        return qp, pp

    @dispatch
    def __call__(
        self, qvec: AbstractPos3D, pvec: AbstractVel
    ) -> tuple[AbstractPos3D, AbstractVel]:
        q = convert(qvec, u.Quantity)
        p = convert(pvec.represent_as(CartesianVel3D, qvec), u.Quantity)

        qp, pp = self(q, p)
        qpvec = CartesianPos3D.from_(qp)
        ppvec = CartesianVel3D.from_(pp)

        return qpvec, ppvec


@dispatch  # type: ignore[misc]
def frame_transform_op(
    from_frame: Galactocentric, to_frame: ICRS, /
) -> _GCF2ICRSOperator:
    """Return a Galactocentric to ICRS frame transformation operator.

    Examples
    --------
    >>> import coordinax.frames as cxf
    >>> icrs_frame = cxf.ICRS()
    >>> gcf_frame = cxf.Galactocentric()
    >>> frame_op = cxf.frame_transform_op(gcf_frame, icrs_frame)
    >>> frame_op
    _GCF2ICRSOperator(
      gcf=Galactocentric(
        galcen=LonLatSphericalPos( ... ),
        roll=Quantity[...](value=weak_i32[], unit=Unit("deg")),
        z_sun=Quantity[...](value=weak_f32[], unit=Unit("pc")),
        galcen_v_sun=CartesianVel3D( ... )
      )
    )

    """
    return _GCF2ICRSOperator(from_frame)



"""Built-in vector classes."""

__all__ = [
    "MathSphericalAcc",
    "MathSphericalPos",
    "MathSphericalVel",
]

from functools import partial
from typing import final
from typing_extensions import override

import equinox as eqx
import jax
from jaxtyping import ArrayLike
from quax import register

import quaxed.lax as qlax
import quaxed.numpy as jnp
from dataclassish import replace
from dataclassish.converters import Unless
from unxt.quantity import AbstractQuantity, Quantity

import coordinax._src.typing as ct
from .base_spherical import (
    AbstractSphericalAcc,
    AbstractSphericalPos,
    AbstractSphericalVel,
    _180d,
    _360d,
)
from coordinax._src.angle import Angle, BatchableAngle
from coordinax._src.distance import AbstractDistance, BatchableDistance, Distance
from coordinax._src.utils import classproperty
from coordinax._src.vectors import checks
from coordinax._src.vectors.converters import converter_azimuth_to_range


@final
class MathSphericalPos(AbstractSphericalPos):
    """Spherical vector representation.

    .. note::

        This class follows the Mathematics conventions.

    Parameters
    ----------
    r : `coordinax.Distance`
        Radial distance r (slant distance to origin),
    theta : `coordinax.angle.Angle`
        Azimuthal angle [0, 360) [deg] where 0 is the x-axis.
    phi : `coordinax.angle.Angle`
        Polar angle [0, 180] [deg] where 0 is the z-axis.

    """

    r: BatchableDistance = eqx.field(
        converter=Unless(AbstractDistance, partial(Distance.from_, dtype=float))
    )
    r"""Radial distance :math:`r \in [0,+\infty)`."""

    theta: BatchableAngle = eqx.field(
        converter=Unless(
            Angle, lambda x: converter_azimuth_to_range(Angle.from_(x, dtype=float))
        )
    )
    r"""Azimuthal angle, generally :math:`\theta \in [0,360)`."""

    phi: BatchableAngle = eqx.field(converter=partial(Angle.from_, dtype=float))
    r"""Inclination angle :math:`\phi \in [0,180]`."""

    def __check_init__(self) -> None:
        """Check the validity of the initialization."""
        checks.check_r_non_negative(self.r)
        checks.check_polar_range(self.phi)

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["MathSphericalVel"]:
        return MathSphericalVel

    @partial(eqx.filter_jit, inline=True)
    def norm(self) -> BatchableDistance:
        """Return the norm of the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx
        >>> s = cx.MathSphericalPos(r=u.Quantity(3, "kpc"),
        ...                         theta=u.Quantity(90, "deg"),
        ...                         phi=u.Quantity(0, "deg"))
        >>> s.norm()
        Distance(Array(3., dtype=float32), unit='kpc')

        """
        return self.r


@MathSphericalPos.from_._f.register  # type: ignore[attr-defined, misc]  # noqa: SLF001
def from_(
    cls: type[MathSphericalPos],
    *,
    r: AbstractQuantity,
    theta: AbstractQuantity,
    phi: AbstractQuantity,
) -> MathSphericalPos:
    """Construct MathSphericalPos, allowing for out-of-range values.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    Let's start with a valid input:

    >>> cx.MathSphericalPos.from_(r=u.Quantity(3, "kpc"),
    ...                           theta=u.Quantity(90, "deg"),
    ...                           phi=u.Quantity(0, "deg"))
    MathSphericalPos(
      r=Distance(value=f32[], unit=Unit("kpc")),
      theta=Angle(value=f32[], unit=Unit("deg")),
      phi=Angle(value=f32[], unit=Unit("deg"))
    )

    The radial distance can be negative, which wraps the azimuthal angle by 180
    degrees and flips the polar angle:

    >>> vec = cx.MathSphericalPos.from_(r=u.Quantity(-3, "kpc"),
    ...                                 theta=u.Quantity(100, "deg"),
    ...                                 phi=u.Quantity(45, "deg"))
    >>> vec.r
    Distance(Array(3., dtype=float32), unit='kpc')
    >>> vec.theta
    Angle(Array(280., dtype=float32), unit='deg')
    >>> vec.phi
    Angle(Array(135., dtype=float32), unit='deg')

    The polar angle can be outside the [0, 180] deg range, causing the azimuthal
    angle to be shifted by 180 degrees:

    >>> vec = cx.MathSphericalPos.from_(r=u.Quantity(3, "kpc"),
    ...                                 theta=u.Quantity(0, "deg"),
    ...                                 phi=u.Quantity(190, "deg"))
    >>> vec.r
    Distance(Array(3., dtype=float32), unit='kpc')
    >>> vec.theta
    Angle(Array(180., dtype=float32), unit='deg')
    >>> vec.phi
    Angle(Array(170., dtype=float32), unit='deg')

    The azimuth can be outside the [0, 360) deg range. This is wrapped to the
    [0, 360) deg range (actually the base constructor does this):

    >>> vec = cx.MathSphericalPos.from_(r=u.Quantity(3, "kpc"),
    ...                                 theta=u.Quantity(365, "deg"),
    ...                                 phi=u.Quantity(90, "deg"))
    >>> vec.theta
    Angle(Array(5., dtype=float32), unit='deg')

    """
    # 1) Convert the inputs
    fields = MathSphericalPos.__dataclass_fields__
    r = fields["r"].metadata["converter"](r)
    theta = fields["theta"].metadata["converter"](theta)
    phi = fields["phi"].metadata["converter"](phi)

    # 2) handle negative distances
    r_pred = r < jnp.zeros_like(r)
    r = qlax.select(r_pred, -r, r)
    theta = qlax.select(r_pred, theta + _180d, theta)
    phi = qlax.select(r_pred, _180d - phi, phi)

    # 3) Handle polar angle outside of [0, 180] degrees
    phi = jnp.mod(phi, _360d)  # wrap to [0, 360) deg
    phi_pred = phi < _180d
    phi = qlax.select(phi_pred, phi, _360d - phi)
    theta = qlax.select(phi_pred, theta, theta + _180d)

    # 4) Construct. This also handles the azimuthal angle wrapping
    return cls(r=r, theta=theta, phi=phi)


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_p_vmsph(lhs: ArrayLike, rhs: MathSphericalPos, /) -> MathSphericalPos:
    """Scale the polar position by a scalar.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> import quaxed.numpy as jnp

    >>> v = cx.MathSphericalPos(r=u.Quantity(3, "kpc"),
    ...                         theta=u.Quantity(90, "deg"),
    ...                         phi=u.Quantity(0, "deg"))

    >>> jnp.linalg.vector_norm(v, axis=-1)
    Quantity['length'](Array(3., dtype=float32), unit='kpc')

    >>> nv = jnp.multiply(2, v)
    >>> nv
    MathSphericalPos(
      r=Distance(value=f32[], unit=Unit("kpc")),
      theta=Angle(value=f32[], unit=Unit("deg")),
      phi=Angle(value=f32[], unit=Unit("deg"))
    )
    >>> nv.r
    Distance(Array(6., dtype=float32), unit='kpc')

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )
    # Scale the radial distance
    return replace(rhs, r=lhs * rhs.r)


##############################################################################


@final
class MathSphericalVel(AbstractSphericalVel):
    """Spherical differential representation."""

    d_r: ct.BatchableSpeed = eqx.field(
        converter=partial(Quantity["speed"].from_, dtype=float)
    )
    r"""Radial speed :math:`dr/dt \in [-\infty, \infty]."""

    d_theta: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(Quantity["angular speed"].from_, dtype=float)
    )
    r"""Azimuthal speed :math:`d\theta/dt \in [-\infty, \infty]."""

    d_phi: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(Quantity["angular speed"].from_, dtype=float)
    )
    r"""Inclination speed :math:`d\phi/dt \in [-\infty, \infty]."""

    @override
    @classproperty
    @classmethod
    def integral_cls(cls) -> type[MathSphericalPos]:
        return MathSphericalPos

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["MathSphericalAcc"]:
        return MathSphericalAcc


##############################################################################


@final
class MathSphericalAcc(AbstractSphericalAcc):
    """Spherical acceleration representation."""

    d2_r: ct.BatchableAcc = eqx.field(
        converter=partial(Quantity["acceleration"].from_, dtype=float)
    )
    r"""Radial acceleration :math:`d^2r/dt^2 \in [-\infty, \infty]."""

    d2_theta: ct.BatchableAngularAcc = eqx.field(
        converter=partial(Quantity["angular acceleration"].from_, dtype=float)
    )
    r"""Azimuthal acceleration :math:`d^2\theta/dt^2 \in [-\infty, \infty]."""

    d2_phi: ct.BatchableAngularAcc = eqx.field(
        converter=partial(Quantity["angular acceleration"].from_, dtype=float)
    )
    r"""Inclination acceleration :math:`d^2\phi/dt^2 \in [-\infty, \infty]."""

    @override
    @classproperty
    @classmethod
    def integral_cls(cls) -> type[MathSphericalVel]:
        return MathSphericalVel



"""Optional dependencies. Internal use only."""

__all__ = ["OptDeps"]

from optional_dependencies import OptionalDependencyEnum, auto


class OptDeps(OptionalDependencyEnum):  # type: ignore[misc]  # pylint: disable=invalid-enum-extension
    """Optional dependencies for ``coordinax_space_frames``."""

    ASTROPY = auto()



"""Representation of coordinates in different systems."""

__all__: list[str] = []


import equinox as eqx

import quaxed.numpy as xp
import unxt as u
from unxt.quantity import AbstractQuantity

from coordinax._src.angle import Angle, BatchableAngleQ

_0d = Angle(0, "rad")
_pid = Angle(180, "deg")
_2pid = Angle(360, "deg")


def check_r_non_negative(r: AbstractQuantity) -> AbstractQuantity:
    """Check that the radial distance is non-negative.

    Examples
    --------
    >>> import unxt as u

    Pass through the input if the radial distance is non-negative.

    >>> x = u.Quantity([0, 1, 2], "m")
    >>> check_r_non_negative(x)
    Quantity['length'](Array([0, 1, 2], dtype=int32), unit='m')

    Raise an error if the radial distance is negative.

    >>> x = u.Quantity([-1, 1, 2], "m")
    >>> try: check_r_non_negative(x)
    ... except Exception: pass

    """
    return check_non_negative(r, name="radial distance r")


def check_polar_range(
    polar: BatchableAngleQ,
    /,
    _l: Angle = _0d,
    _u: Angle = _pid,
) -> BatchableAngleQ:
    """Check that the polar angle is in the range.

    Examples
    --------
    >>> import unxt as u
    >>> from coordinax._src.vectors.checks import check_polar_range

    Pass through the input if it's in the range.

    >>> x = u.Quantity([0., 1, 2], "deg")
    >>> check_polar_range(x)
    Quantity['angle'](Array([0., 1., 2.], dtype=float32), unit='deg')

    Raise an error if anything is outside the range.

    >>> x = u.Quantity([0., 1, 2], "m")
    >>> try: check_polar_range(x)
    ... except Exception as e: print("wrong units")
    wrong units

    >>> x = u.Quantity([-1., 1, 2], "deg")
    >>> try: check_polar_range(x)
    ... except Exception: pass

    """
    polar = eqx.error_if(
        polar,
        not u.is_unit_convertible("deg", polar),
        "The polar angle must be in angular units.",
    )
    return eqx.error_if(
        polar,
        xp.any(xp.logical_or((polar < _l), (polar > _u))),
        "The inclination angle must be in the range [0, pi].",
    )


def check_non_negative(x: AbstractQuantity, /, *, name: str = "") -> AbstractQuantity:
    """Check that the input is non-negative.

    Examples
    --------
    >>> from unxt import Quantity

    Pass through the input if the value is non-negative.

    >>> x = Quantity([0, 1, 2], "m")
    >>> check_non_negative(x)
    Quantity['length'](Array([0, 1, 2], dtype=int32), unit='m')

    Raise an error if any value is negative.

    >>> x = Quantity([-1, 1, 2], "m")
    >>> try: check_non_negative(x)
    ... except Exception: pass

    """
    name = f" {name}" if name else name
    return eqx.error_if(x, xp.any(x < 0), f"The input{name} must be non-negative.")


def check_non_negative_non_zero(
    x: AbstractQuantity, /, *, name: str = ""
) -> AbstractQuantity:
    """Check that the input is non-negative and non-zero.

    Examples
    --------
    >>> from unxt import Quantity

    Pass through the input if the value is non-negative.

    >>> x = Quantity([1, 2, 3], "m")
    >>> check_non_negative_non_zero(x)
    Quantity['length'](Array([1, 2, 3], dtype=int32), unit='m')

    Raise an error if any value is negative or zero.

    >>> x = Quantity([-1, 1, 2], "m")
    >>> try: check_non_negative_non_zero(x)
    ... except Exception: pass

    >>> x = Quantity([0, 1, 2], "m")
    >>> try: check_non_negative_non_zero(x)
    ... except Exception: pass

    """
    name = f" {name}" if name else name
    return eqx.error_if(
        x, xp.any(x <= 0), f"The input{name} must be non-negative and non-zero."
    )


def check_less_than(
    x: AbstractQuantity,
    max_val: AbstractQuantity,
    /,
    *,
    name: str = "",
    comparison_name: str = "the specified maximum value",
) -> AbstractQuantity:
    """Check that the input value is less than the input maximum value.

    Examples
    --------
    >>> from unxt import Quantity

    Pass through the input if the value is less than the max value:

    >>> x = Quantity([1, 2, 3], "m")
    >>> check_less_than(x, Quantity(4, "m"))
    Quantity['length'](Array([1, 2, 3], dtype=int32), unit='m')

    Raise an error if the input is larger than the maximum value.

    >>> try: check_less_than(x, Quantity(1.5, "m"))
    ... except Exception: pass

    """
    name = f" {name}" if name else name
    msg = f"The input{name} must be less than {comparison_name}."
    return eqx.error_if(x, xp.any(x >= max_val), msg)


def check_less_than_equal(
    x: AbstractQuantity,
    max_val: AbstractQuantity,
    /,
    *,
    name: str = "",
    comparison_name: str = "the specified maximum value",
) -> AbstractQuantity:
    """Check that the input value is less than or equal to the input maximum value.

    Examples
    --------
    >>> from unxt import Quantity

    Pass through the input if the value is less than or equal to the max value:

    >>> x = Quantity([1, 2, 3], "m")
    >>> check_less_than_equal(x, Quantity(3, "m"))
    Quantity['length'](Array([1, 2, 3], dtype=int32), unit='m')

    Raise an error if the input is larger than the maximum value.

    >>> try: check_less_than_equal(x, Quantity(2, "m"))
    ... except Exception: pass

    """
    name = f" {name}" if name else name
    msg = f"The input{name} must be less than or equal to {comparison_name}."
    return eqx.error_if(x, xp.any(x > max_val), msg)


def check_greater_than(
    x: AbstractQuantity,
    min_val: AbstractQuantity,
    /,
    *,
    name: str = "",
    comparison_name: str = "the specified minimum value",
) -> AbstractQuantity:
    """Check that the input value is greater than the input minimum value.

    Examples
    --------
    >>> from unxt import Quantity

    Pass through the input if the value is greater than the min value:

    >>> x = Quantity([1, 2, 3], "m")
    >>> check_greater_than(x, Quantity(0, "m"))
    Quantity['length'](Array([1, 2, 3], dtype=int32), unit='m')

    Raise an error if the input is smaller than the minimum value.

    >>> try: check_greater_than(x, Quantity(4, "m"))
    ... except Exception: pass

    """
    name = f" {name}" if name else name
    msg = f"The input{name} must be greater than {comparison_name}."
    return eqx.error_if(x, xp.any(x <= min_val), msg)


def check_greater_than_equal(
    x: AbstractQuantity,
    min_val: AbstractQuantity,
    /,
    *,
    name: str = "",
    comparison_name: str = "the specified minimum value",
) -> AbstractQuantity:
    """Check that the input value is greater than or equal to the input minimum value.

    Examples
    --------
    >>> from unxt import Quantity

    Pass through the input if the value is greater than or equal to the min value:

    >>> x = Quantity([1, 2, 3], "m")
    >>> check_greater_than_equal(x, Quantity(1, "m"))
    Quantity['length'](Array([1, 2, 3], dtype=int32), unit='m')

    Raise an error if the input is smaller than the minimum value.

    >>> try: check_greater_than_equal(x, Quantity(2, "m"))
    ... except Exception: pass

    """
    name = f" {name}" if name else name
    msg = f"The input{name} must be greater than or equal to {comparison_name}."
    return eqx.error_if(x, xp.any(x < min_val), msg)



"""Package setup information.

Note that this module is NOT public API nor are any of its contents.
Stability is NOT guaranteed.
This module exposes package setup information for the `unxt` package.

"""

__all__: list[str] = []

import os
from typing import Final

RUNTIME_TYPECHECKER: Final[str | None] = (
    v
    if (v := os.environ.get("COORDINAX_ENABLE_RUNTIME_TYPECHECKING", None)) != "None"
    else None
)
"""Runtime type checking variable "COORDINAX_ENABLE_RUNTIME_TYPECHECKING".

Set to "None" to disable runtime typechecking (default). Set to
"beartype.beartype" to enable runtime typechecking.

See https://docs.kidger.site/jaxtyping/api/runtime-type-checking for more
information on options.

"""



"""Transformations between representations."""

__all__: list[str] = []

from dataclasses import replace
from functools import partial
from math import prod
from typing import Any

import equinox as eqx
import jax
from plum import dispatch

import quaxed.numpy as jnp
import unxt as u
from dataclassish import field_items

from coordinax._src.distance import AbstractDistance
from coordinax._src.vectors.base import AbstractPos, AbstractVel
from coordinax._src.vectors.base.flags import AttrFilter
from coordinax._src.vectors.d1 import AbstractVel1D
from coordinax._src.vectors.d2 import AbstractVel2D
from coordinax._src.vectors.d3 import AbstractVel3D


# TODO: implement for cross-representations
@dispatch.multi(  # type: ignore[misc]
    # N-D -> N-D
    (AbstractVel1D, type[AbstractVel1D], AbstractPos | u.Quantity["length"]),
    (AbstractVel2D, type[AbstractVel2D], AbstractPos | u.Quantity["length"]),
    (AbstractVel3D, type[AbstractVel3D], AbstractPos | u.Quantity["length"]),
)
def represent_as(
    current: AbstractVel,
    target: type[AbstractVel],
    position: AbstractPos | u.Quantity["length"],
    /,
    **kwargs: Any,
) -> AbstractVel:
    """AbstractVel -> Cartesian -> AbstractVel.

    This is the base case for the transformation of vector differentials.

    Parameters
    ----------
    current : AbstractVel
        The vector differential to transform.
    target : type[AbstractVel]
        The target type of the vector differential.
    position : AbstractPos
        The position vector used to transform the differential.
    **kwargs : Any
        Additional keyword arguments.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    Let's start in 1D:

    >>> q = cx.CartesianPos1D.from_(1.0, "km")
    >>> p = cx.CartesianVel1D.from_(1.0, "km/s")
    >>> cx.represent_as(p, cx.RadialVel, q)
    RadialVel( d_r=Quantity[...]( value=f32[], unit=Unit("km / s") ) )

    Now in 2D:

    >>> q = cx.CartesianPos2D.from_([1.0, 2.0], "km")
    >>> p = cx.CartesianVel2D.from_([1.0, 2.0], "km/s")
    >>> cx.represent_as(p, cx.PolarVel, q)
    PolarVel(
      d_r=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_phi=Quantity[...]( value=f32[], unit=Unit("rad / s") )
    )

    And in 3D:

    >>> q = cx.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")
    >>> p = cx.CartesianVel3D.from_([1.0, 2.0, 3.0], "km/s")
    >>> cx.represent_as(p, cx.SphericalVel, q)
    SphericalVel(
      d_r=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_theta=Quantity[...]( value=f32[], unit=Unit("rad / s") ),
      d_phi=Quantity[...]( value=f32[], unit=Unit("rad / s") )
    )

    If given a position as a Quantity, it will be converted to the appropriate
    Cartesian vector:

    >>> p = cx.CartesianVel3D.from_([1.0, 2.0, 3.0], "km/s")
    >>> cx.represent_as(p, cx.SphericalVel, u.Quantity([1.0, 2.0, 3.0], "km"))
    SphericalVel(
      d_r=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_theta=Quantity[...]( value=f32[], unit=Unit("rad / s") ),
      d_phi=Quantity[...]( value=f32[], unit=Unit("rad / s") )
    )

    """
    # TODO: not require the shape munging / support more shapes
    shape = current.shape
    flat_shape = prod(shape)

    # Parse the position to an AbstractPos
    if isinstance(position, AbstractPos):
        posvec = position
    else:  # Q -> Cart<X>D
        posvec = current.integral_cls._cartesian_cls.from_(  # noqa: SLF001
            position
        )

    posvec = posvec.reshape(flat_shape)  # flattened

    # Start by transforming the position to the type required by the
    # differential to construct the Jacobian.
    current_pos = represent_as(posvec, current.integral_cls, **kwargs)
    # TODO: not need to cast to distance
    current_pos = replace(
        current_pos,
        **{
            k: v.distance
            for k, v in field_items(AttrFilter, current_pos)
            if isinstance(v, AbstractDistance)
        },
    )

    # Takes the Jacobian through the representation transformation function.  This
    # returns a representation of the target type, where the value of each field the
    # corresponding row of the Jacobian. The value of the field is a Quantity with
    # the correct numerator unit (of the Jacobian row). The value is a Vector of the
    # original type, with fields that are the columns of that row, but with only the
    # denomicator's units.
    tmp = partial(represent_as, **kwargs)
    jac_rep_as = eqx.filter_jit(jax.vmap(jax.jacfwd(tmp), in_axes=(0, None)))
    jac_nested_vecs = jac_rep_as(current_pos, target.integral_cls)

    # This changes the Jacobian to be a dictionary of each row, with the value
    # being that row's column as a dictionary, now with the correct units for
    # each element:  {row_i: {col_j: Quantity(value, row.unit / column.unit)}}
    jac_rows = {
        f"d_{k}": {
            kk: u.Quantity(vv.value, unit=v.unit / vv.unit)
            for kk, vv in field_items(AttrFilter, v.value)
        }
        for k, v in field_items(AttrFilter, jac_nested_vecs)
    }

    # Now we can use the Jacobian to transform the differential.
    flat_current = current.reshape(flat_shape)
    newvec = target(
        **{  # Each field is the dot product of the row of the J and the diff column.
            k: jnp.sum(  # Doing the dot product.
                jnp.stack(
                    tuple(
                        j_c * getattr(flat_current, f"d_{kk}")
                        for kk, j_c in j_r.items()
                    )
                ),
                axis=0,
            )
            for k, j_r in jac_rows.items()
        }
    ).reshape(shape)

    # TODO: add  df(q)/dt, which is 0 for all current transforms

    return newvec  # noqa: RET504



"""Compatibility for Quantity."""

__all__: list[str] = []

from plum import add_promotion_rule, conversion_method

from unxt.quantity import AbstractQuantity, Quantity

from .base import AbstractAngle
from .core import Angle

# Add a rule that when a AbstractAngle interacts with a Quantity, the
# angle degrades to a Quantity. This is necessary for many operations, e.g.
# division of an angle by non-dimensionless quantity where the resulting units
# are not those of an angle.
add_promotion_rule(AbstractAngle, Quantity, Quantity)


@conversion_method(type_from=AbstractAngle, type_to=Quantity)  # type: ignore[misc]
def _convert_angle_to_quantity(x: AbstractAngle) -> Quantity:
    """Convert a distance to a quantity.

    Examples
    --------
    >>> import unxt as u
    >>> from coordinax.angle import Angle
    >>> from plum import convert

    >>> a = Angle(90, "deg")
    >>> convert(a, u.Quantity)
    Quantity['angle'](Array(90, dtype=int32, weak_type=True), unit='deg')

    """
    return Quantity(x.value, x.unit)


@conversion_method(type_from=AbstractQuantity, type_to=Angle)  # type: ignore[misc]
def _convert_quantity_to_angle(q: AbstractQuantity, /) -> Angle:
    """Convert any quantity to an Angle.

    Examples
    --------
    >>> from plum import convert
    >>> from unxt.quantity import Quantity, UncheckedQuantity
    >>> from coordinax.angle import Angle
    >>> q = UncheckedQuantity(1, "rad")
    >>> q
    UncheckedQuantity(Array(1, dtype=int32, ...), unit='rad')

    >>> convert(q, Angle)
    Angle(Array(1, dtype=int32, ...), unit='rad')

    The self-conversion doesn't copy the object:

    >>> q = Angle(1, "rad")
    >>> convert(q, Angle) is q
    True

    """
    if isinstance(q, Angle):
        return q
    return Angle(q.value, q.unit)



"""Built-in vector classes."""

__all__ = [
    "CartesianAcc3D",
    "CartesianPos3D",
    "CartesianVel3D",
]

from dataclasses import fields, replace
from functools import partial
from typing import final
from typing_extensions import override

import equinox as eqx
import jax
from jaxtyping import ArrayLike, Shaped
from plum import dispatch
from quax import register

import quaxed.lax as qlax
import quaxed.numpy as jnp
import unxt as u
from dataclassish import field_items
from unxt.quantity import AbstractQuantity

import coordinax._src.typing as ct
from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D
from .generic import CartesianGeneric3D
from coordinax._src.distance import BatchableLength
from coordinax._src.utils import classproperty
from coordinax._src.vectors.base import AbstractPos
from coordinax._src.vectors.base.mixins import AvalMixin

#####################################################################
# Position


@final
class CartesianPos3D(AbstractPos3D):
    """Cartesian vector representation."""

    x: BatchableLength = eqx.field(
        converter=partial(u.Quantity["length"].from_, dtype=float)
    )
    r"""X coordinate :math:`x \in (-\infty,+\infty)`."""

    y: BatchableLength = eqx.field(
        converter=partial(u.Quantity["length"].from_, dtype=float)
    )
    r"""Y coordinate :math:`y \in (-\infty,+\infty)`."""

    z: BatchableLength = eqx.field(
        converter=partial(u.Quantity["length"].from_, dtype=float)
    )
    r"""Z coordinate :math:`z \in (-\infty,+\infty)`."""

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["CartesianVel3D"]:
        """Return the differential of the class."""
        return CartesianVel3D


# =====================================================
# Constructors


@CartesianPos3D.from_.dispatch  # type: ignore[attr-defined, misc]
def from_(
    cls: type[CartesianPos3D],
    obj: AbstractQuantity,  # TODO: Shaped[AbstractQuantity, "*batch 3"]
    /,
) -> CartesianPos3D:
    """Construct a 3D Cartesian position.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_(u.Quantity([1, 2, 3], "m"))
    >>> vec
    CartesianPos3D(
      x=Quantity[...](value=f32[], unit=Unit("m")),
      y=Quantity[...](value=f32[], unit=Unit("m")),
      z=Quantity[...](value=f32[], unit=Unit("m"))
    )

    """
    comps = {f.name: obj[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)


# =====================================================
# Primitives


@register(jax.lax.add_p)  # type: ignore[misc]
def _add_cart3d_pos(lhs: CartesianPos3D, rhs: AbstractPos, /) -> CartesianPos3D:
    """Subtract two vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> s = cx.SphericalPos(r=u.Quantity(1, "kpc"), theta=u.Quantity(90, "deg"),
    ...                     phi=u.Quantity(0, "deg"))
    >>> (q + s).x
    Quantity['length'](Array(2., dtype=float32), unit='kpc')

    """
    cart = rhs.represent_as(CartesianPos3D)
    return replace(
        lhs, **{k: qlax.add(v, getattr(cart, k)) for k, v in field_items(lhs)}
    )


@register(jax.lax.neg_p)  # type: ignore[misc]
def _neg_p_cart3d_pos(obj: CartesianPos3D, /) -> CartesianPos3D:
    """Negate the `coordinax.CartesianPos3D`.

    Examples
    --------
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> (-q).x
    Quantity['length'](Array(-1., dtype=float32), unit='kpc')

    """
    return jax.tree.map(qlax.neg, obj)


@register(jax.lax.sub_p)  # type: ignore[misc]
def _sub_cart3d_pos(lhs: CartesianPos3D, rhs: AbstractPos, /) -> CartesianPos3D:
    """Subtract two vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> s = cx.SphericalPos(r=u.Quantity(1, "kpc"), theta=u.Quantity(90, "deg"),
    ...                     phi=u.Quantity(0, "deg"))
    >>> (q - s).x
    Quantity['length'](Array(0., dtype=float32), unit='kpc')

    """
    cart = rhs.represent_as(CartesianPos3D)
    return jax.tree.map(qlax.sub, lhs, cart)


# =====================================================
# Functions


# from coordinax.vectors.funcs
@dispatch  # type: ignore[misc]
@partial(eqx.filter_jit, inline=True)
def normalize_vector(obj: CartesianPos3D, /) -> CartesianGeneric3D:
    """Return the norm of the vector.

    This has length 1.

    .. note::

        The unit vector is dimensionless, even if the input vector has units.
        This is because the unit vector is a ratio of two quantities: each
        component and the norm of the vector.

    Returns
    -------
    CartesianGeneric3D
        The norm of the vector.

    Examples
    --------
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> cx.normalize_vector(q)
    CartesianGeneric3D(
      x=Quantity[...]( value=f32[], unit=Unit(dimensionless) ),
      y=Quantity[...]( value=f32[], unit=Unit(dimensionless) ),
      z=Quantity[...]( value=f32[], unit=Unit(dimensionless) )
    )

    """
    norm = obj.norm()
    return CartesianGeneric3D(x=obj.x / norm, y=obj.y / norm, z=obj.z / norm)


#####################################################################
# Velocity


@final
class CartesianVel3D(AvalMixin, AbstractVel3D):
    """Cartesian differential representation."""

    d_x: ct.BatchableSpeed = eqx.field(
        converter=partial(u.Quantity["speed"].from_, dtype=float)
    )
    r"""X speed :math:`dx/dt \in [-\infty, \infty]."""

    d_y: ct.BatchableSpeed = eqx.field(
        converter=partial(u.Quantity["speed"].from_, dtype=float)
    )
    r"""Y speed :math:`dy/dt \in [-\infty, \infty]."""

    d_z: ct.BatchableSpeed = eqx.field(
        converter=partial(u.Quantity["speed"].from_, dtype=float)
    )
    r"""Z speed :math:`dz/dt \in [-\infty, \infty]."""

    @override
    @classproperty
    @classmethod
    def integral_cls(cls) -> type[CartesianPos3D]:
        return CartesianPos3D

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["CartesianAcc3D"]:
        return CartesianAcc3D

    @partial(eqx.filter_jit, inline=True)
    def norm(self, _: AbstractPos3D | None = None, /) -> ct.BatchableSpeed:
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> c = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
        >>> c.norm()
        Quantity['speed'](Array(3.7416575, dtype=float32), unit='km / s')

        """
        return jnp.sqrt(self.d_x**2 + self.d_y**2 + self.d_z**2)


# -----------------------------------------------------


@CartesianVel3D.from_.dispatch  # type: ignore[attr-defined,misc]
def from_(
    cls: type[CartesianVel3D],
    obj: AbstractQuantity,  # TODO: Shaped[AbstractQuantity, "*batch 3"]
    /,
) -> CartesianVel3D:
    """Construct a 3D Cartesian velocity.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CartesianVel3D.from_(u.Quantity([1, 2, 3], "m/s"))
    >>> vec
    CartesianVel3D(
      d_x=Quantity[...]( value=f32[], unit=Unit("m / s") ),
      d_y=Quantity[...]( value=f32[], unit=Unit("m / s") ),
      d_z=Quantity[...]( value=f32[], unit=Unit("m / s") )
    )

    """
    comps = {f.name: obj[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)


# -----------------------------------------------------
# Method dispatches


@register(jax.lax.add_p)  # type: ignore[misc]
def _add_pp(lhs: CartesianVel3D, rhs: CartesianVel3D, /) -> CartesianVel3D:
    """Add two Cartesian velocities.

    Examples
    --------
    >>> import coordinax as cx
    >>> q = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> q2 = q + q
    >>> q2.d_y
    Quantity['speed'](Array(4., dtype=float32), unit='km / s')

    """
    return jax.tree.map(qlax.add, lhs, rhs)


@register(jax.lax.sub_p)  # type: ignore[misc]
def _sub_v3_v3(lhs: CartesianVel3D, other: CartesianVel3D, /) -> CartesianVel3D:
    """Subtract two differentials.

    Examples
    --------
    >>> from coordinax import CartesianPos3D, CartesianVel3D
    >>> q = CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> q2 = q - q
    >>> q2.d_y
    Quantity['speed'](Array(0., dtype=float32), unit='km / s')

    """
    return jax.tree.map(qlax.sub, lhs, other)


#####################################################################
# Acceleration


@final
class CartesianAcc3D(AvalMixin, AbstractAcc3D):
    """Cartesian differential representation."""

    d2_x: ct.BatchableAcc = eqx.field(
        converter=partial(u.Quantity["acceleration"].from_, dtype=float)
    )
    r"""X acceleration :math:`d^2x/dt^2 \in [-\infty, \infty]."""

    d2_y: ct.BatchableAcc = eqx.field(
        converter=partial(u.Quantity["acceleration"].from_, dtype=float)
    )
    r"""Y acceleration :math:`d^2y/dt^2 \in [-\infty, \infty]."""

    d2_z: ct.BatchableAcc = eqx.field(
        converter=partial(u.Quantity["acceleration"].from_, dtype=float)
    )
    r"""Z acceleration :math:`d^2z/dt^2 \in [-\infty, \infty]."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[CartesianVel3D]:
        return CartesianVel3D

    # -----------------------------------------------------
    # Methods

    @override
    @partial(eqx.filter_jit, inline=True)
    def norm(self, _: AbstractVel3D | None = None, /) -> ct.BatchableAcc:
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> c = cx.CartesianAcc3D.from_([1, 2, 3], "km/s2")
        >>> c.norm()
        Quantity['acceleration'](Array(3.7416575, dtype=float32), unit='km / s2')

        """
        return jnp.sqrt(self.d2_x**2 + self.d2_y**2 + self.d2_z**2)


# -----------------------------------------------------


@CartesianAcc3D.from_.dispatch  # type: ignore[attr-defined, misc]
def from_(
    cls: type[CartesianAcc3D], obj: Shaped[AbstractQuantity, "*batch 3"], /
) -> CartesianAcc3D:
    """Construct a 3D Cartesian acceleration.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CartesianAcc3D.from_(u.Quantity([1, 2, 3], "m/s2"))
    >>> vec
    CartesianAcc3D(
      d2_x=Quantity[...](value=f32[], unit=Unit("m / s2")),
      d2_y=Quantity[...](value=f32[], unit=Unit("m / s2")),
      d2_z=Quantity[...](value=f32[], unit=Unit("m / s2"))
    )

    """
    comps = {f.name: obj[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)


# -----------------------------------------------------
# Method dispatches


@register(jax.lax.add_p)  # type: ignore[misc]
def _add_aa(lhs: CartesianAcc3D, rhs: CartesianAcc3D, /) -> CartesianAcc3D:
    """Add two Cartesian accelerations."""
    return jax.tree.map(qlax.add, lhs, rhs)


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_ac3(lhs: ArrayLike, rhs: CartesianPos3D, /) -> CartesianPos3D:
    """Scale a position by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    >>> v = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> jnp.multiply(2, v).x
    Quantity['length'](Array(2., dtype=float32), unit='kpc')

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )

    # Scale the components
    return replace(rhs, x=lhs * rhs.x, y=lhs * rhs.y, z=lhs * rhs.z)


@register(jax.lax.sub_p)  # type: ignore[misc]
def _sub_a3_a3(lhs: CartesianAcc3D, rhs: CartesianAcc3D, /) -> CartesianAcc3D:
    """Subtract two accelerations."""
    return jax.tree.map(qlax.sub, lhs, rhs)



"""Register Quantity support for jax primitives."""
# pylint: disable=import-error, too-many-lines

__all__: list[str] = []

from collections.abc import Callable
from typing import Any, TypeVar

from jax import lax
from jax.core import Primitive
from jaxtyping import ArrayLike
from quax import register as register_

import unxt as u
from quaxed import lax as qlax

from .base import AbstractAngle

T = TypeVar("T")

one = u.unit("")
radian = u.unit("radian")


def register(primitive: Primitive, **kwargs: Any) -> Callable[[T], T]:
    """`quax.register`, but makes mypy happy."""
    return register_(primitive, **kwargs)


################################################################################
# Registering Primitives


# TODO: can this be done with promotion/conversion instead?
@register(lax.cbrt_p)
def _cbrt_p_a(x: AbstractAngle) -> u.Quantity:
    """Cube root of an angle.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from coordinax.angle import Angle

    >>> q = Angle(8, "rad")
    >>> jnp.cbrt(q)
    Quantity['rad1/3'](Array(2., dtype=float32, weak_type=True), unit='rad(1/3)')

    """
    return u.Quantity(lax.cbrt(x.value), unit=x.unit ** (1 / 3))


# ==============================================================================


@register(lax.cos_p)
def _cos_p(x: AbstractAngle) -> u.Quantity:
    """Cosine of an Angle.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from coordinax.angle import Angle

    >>> q = Angle(0, "deg")
    >>> jnp.cos(q)
    Quantity['dimensionless'](Array(1., dtype=float32, ...), unit='')

    """
    return u.Quantity(qlax.cos(u.ustrip(radian, x)), unit=one)


# ==============================================================================


@register(lax.dot_general_p)
def _dot_general_aa(
    lhs: AbstractAngle, rhs: AbstractAngle, /, **kwargs: Any
) -> u.Quantity:
    """Dot product of two Angles.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from coordinax.angle import Angle

    >>> q1 = Angle([1, 2, 3], "deg")
    >>> q2 = Angle([4, 5, 6], "deg")
    >>> jnp.vecdot(q1, q2)
    Quantity['solid angle'](Array(32, dtype=int32), unit='deg2')

    >>> q1 @ q2
    Quantity['solid angle'](Array(32, dtype=int32), unit='deg2')

    """
    return u.Quantity(
        lax.dot_general_p.bind(lhs.value, rhs.value, **kwargs),
        unit=lhs.unit * rhs.unit,
    )


# ==============================================================================


@register(lax.integer_pow_p)
def _integer_pow_p_a(x: AbstractAngle, *, y: Any) -> u.Quantity:
    """Integer power of an Angle.

    Examples
    --------
    >>> from coordinax.angle import Angle
    >>> q = Angle(2, "deg")

    >>> q ** 3
    Quantity['rad3'](Array(8, dtype=int32, weak_type=True), unit='deg3')

    """
    return u.Quantity(value=lax.integer_pow(x.value, y), unit=x.unit**y)


# ==============================================================================


@register(lax.pow_p)
def _pow_p_a(x: AbstractAngle, y: ArrayLike) -> u.Quantity:
    """Power of an Angle by redispatching to Quantity.

    Examples
    --------
    >>> import math
    >>> from coordinax.angle import Angle

    >>> q1 = Angle(10.0, "deg")
    >>> y = 3.0
    >>> q1 ** y
    Quantity['rad3'](Array(1000., dtype=float32, ...), unit='deg3')

    """
    return u.Quantity(x.value, x.unit) ** y  # TODO: better call to power


# ==============================================================================


@register(lax.sin_p)
def _sin_p(x: AbstractAngle) -> u.Quantity:
    """Sine of an Angle.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from coordinax.angle import Angle

    >>> q = Angle(90, "deg")
    >>> jnp.sin(q)
    Quantity['dimensionless'](Array(1., dtype=float32, ...), unit='')

    """
    return u.Quantity(qlax.sin(u.ustrip(radian, x)), unit=one)


# ==============================================================================


@register(lax.sqrt_p)
def _sqrt_p_a(x: AbstractAngle) -> u.Quantity:
    """Square root of an Angle.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from coordinax.angle import Angle

    >>> q = Angle(9, "deg")
    >>> jnp.sqrt(q)
    Quantity['rad0.5'](Array(3., dtype=float32, ...), unit='deg(1/2)')

    """
    # Promote to something that supports sqrt units.
    return u.Quantity(lax.sqrt(x.value), unit=x.unit ** (1 / 2))



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from unxt.quantity import AbstractQuantity

from coordinax._src.angle import Angle

_2pid = Angle(360, "deg")


def converter_azimuth_to_range(
    phi: AbstractQuantity,
) -> AbstractQuantity:
    """Wrap a polar angle to the range [0, 2pi).

    It's safe to do this conversion since this is a phase cut, unlike `theta`,
    which is only on half the sphere.

    Examples
    --------
    >>> import unxt as u
    >>> x = u.Quantity(370, "deg")
    >>> converter_azimuth_to_range(x)
    Quantity['angle'](Array(10, dtype=int32, ...), unit='deg')

    """
    return phi % _2pid



"""Transformations between representations."""

__all__: list[str] = []

from typing import Any
from warnings import warn

from plum import dispatch

import quaxed.numpy as jnp

from coordinax._src.vectors.d1 import AbstractPos1D, CartesianPos1D, RadialPos
from coordinax._src.vectors.d2 import AbstractPos2D, CartesianPos2D, PolarPos
from coordinax._src.vectors.d3 import (
    CartesianPos3D,
    CylindricalPos,
    MathSphericalPos,
    ProlateSpheroidalPos,
    SphericalPos,
)
from coordinax._src.vectors.exceptions import IrreversibleDimensionChange

# =============================================================================
# CartesianPos3D


# -----------------------------------------------
# 1D


@dispatch
def represent_as(
    current: CartesianPos3D, target: type[CartesianPos1D], /, **kwargs: Any
) -> CartesianPos1D:
    """CartesianPos3D -> CartesianPos1D.

    The `y` and `z` coordinates are dropped.

    Examples
    --------
    >>> import warnings
    >>> import coordinax as cx

    >>> x = cx.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos1D)
    >>> x2
    CartesianPos1D(x=Quantity[...](value=f32[], unit=Unit("km")))

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.x)


@dispatch
def represent_as(
    current: CartesianPos3D, target: type[RadialPos], /, **kwargs: Any
) -> RadialPos:
    """CartesianPos3D -> RadialPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.RadialPos)
    >>> x2
    RadialPos(r=Distance(value=f32[], unit=Unit("km")))

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=jnp.sqrt(current.x**2 + current.y**2 + current.z**2))


# -----------------------------------------------
# 2D


@dispatch
def represent_as(
    current: CartesianPos3D, target: type[CartesianPos2D], /, **kwargs: Any
) -> CartesianPos2D:
    """CartesianPos3D -> CartesianPos2D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos2D)
    >>> x2
    CartesianPos2D( x=Quantity[...](value=f32[], unit=Unit("km")),
                    y=Quantity[...](value=f32[], unit=Unit("km")) )

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.x, y=current.y)


@dispatch.multi(
    (CartesianPos3D, type[PolarPos]),
)
def represent_as(
    current: CartesianPos3D, target: type[AbstractPos2D], /, **kwargs: Any
) -> AbstractPos2D:
    """CartesianPos3D -> Cartesian2D -> AbstractPos2D.

    Examples
    --------
    >>> import warnings
    >>> import coordinax as cx

    >>> x = cx.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.PolarPos)
    >>> x2
    PolarPos( r=Distance(value=f32[], unit=Unit("km")),
              phi=Angle(value=f32[], unit=Unit("rad")) )

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    cart2 = represent_as(current, CartesianPos2D)
    return represent_as(cart2, target)


# =============================================================================
# CylindricalPos


# -----------------------------------------------
# 1D


@dispatch
def represent_as(
    current: CylindricalPos, target: type[CartesianPos1D], /, **kwargs: Any
) -> CartesianPos1D:
    """CylindricalPos -> CartesianPos1D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CylindricalPos(rho=u.Quantity(1.0, "km"), phi=u.Quantity(10.0, "deg"),
    ...                       z=u.Quantity(14, "km"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos1D)
    >>> x2
    CartesianPos1D(x=Quantity[...](value=f32[], unit=Unit("km")))

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.rho * jnp.cos(current.phi))


@dispatch
def represent_as(
    current: CylindricalPos, target: type[RadialPos], /, **kwargs: Any
) -> RadialPos:
    """CylindricalPos -> RadialPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CylindricalPos(rho=u.Quantity(1.0, "km"), phi=u.Quantity(10.0, "deg"),
    ...                        z=u.Quantity(14, "km"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.RadialPos)
    >>> x2
    RadialPos(r=Distance(value=f32[], unit=Unit("km")))

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=current.rho)


# -----------------------------------------------
# 2D


@dispatch
def represent_as(
    current: CylindricalPos, target: type[CartesianPos2D], /, **kwargs: Any
) -> CartesianPos2D:
    """CylindricalPos -> CartesianPos2D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CylindricalPos(rho=u.Quantity(1.0, "km"), phi=u.Quantity(10.0, "deg"),
    ...                       z=u.Quantity(14, "km"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos2D)
    >>> x2
    CartesianPos2D( x=Quantity[...](value=f32[], unit=Unit("km")),
                    y=Quantity[...](value=f32[], unit=Unit("km")) )

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    x = current.rho * jnp.cos(current.phi)
    y = current.rho * jnp.sin(current.phi)
    return target(x=x, y=y)


@dispatch
def represent_as(
    current: CylindricalPos, target: type[PolarPos], /, **kwargs: Any
) -> PolarPos:
    """CylindricalPos -> PolarPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CylindricalPos(rho=u.Quantity(1.0, "km"), phi=u.Quantity(10.0, "deg"),
    ...                       z=u.Quantity(14, "km"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.PolarPos)
    >>> x2
    PolarPos( r=Distance(value=f32[], unit=Unit("km")),
              phi=Angle(value=f32[], unit=Unit("deg")) )

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=current.rho, phi=current.phi)


# =============================================================================
# SphericalPos


# -----------------------------------------------
# 1D


@dispatch
def represent_as(
    current: SphericalPos, target: type[CartesianPos1D], /, **kwargs: Any
) -> CartesianPos1D:
    """SphericalPos -> CartesianPos1D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.SphericalPos(r=u.Quantity(1.0, "km"), theta=u.Quantity(14, "deg"),
    ...                     phi=u.Quantity(10.0, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos1D)
    >>> x2
    CartesianPos1D(x=Quantity[...](value=f32[], unit=Unit("km")))

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.r * jnp.sin(current.theta) * jnp.cos(current.phi))


@dispatch
def represent_as(
    current: SphericalPos, target: type[RadialPos], /, **kwargs: Any
) -> RadialPos:
    """SphericalPos -> RadialPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.SphericalPos(r=u.Quantity(1.0, "km"), theta=u.Quantity(14, "deg"),
    ...                     phi=u.Quantity(10.0, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.RadialPos)
    >>> x2
    RadialPos(r=Distance(value=f32[], unit=Unit("km")))

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=current.r)


# -----------------------------------------------
# 2D


@dispatch
def represent_as(
    current: SphericalPos, target: type[CartesianPos2D], /, **kwargs: Any
) -> CartesianPos2D:
    """SphericalPos -> CartesianPos2D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.SphericalPos(r=u.Quantity(1.0, "km"), theta=u.Quantity(14, "deg"),
    ...                     phi=u.Quantity(10.0, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos2D)
    >>> x2
    CartesianPos2D( x=Quantity[...](value=f32[], unit=Unit("km")),
                    y=Quantity[...](value=f32[], unit=Unit("km")) )

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    x = current.r * jnp.sin(current.theta) * jnp.cos(current.phi)
    y = current.r * jnp.sin(current.theta) * jnp.sin(current.phi)
    return target(x=x, y=y)


@dispatch
def represent_as(
    current: SphericalPos, target: type[PolarPos], /, **kwargs: Any
) -> PolarPos:
    """SphericalPos -> PolarPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.SphericalPos(r=u.Quantity(1.0, "km"), theta=u.Quantity(14, "deg"),
    ...                     phi=u.Quantity(10.0, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.PolarPos)
    >>> x2
    PolarPos( r=Distance(value=f32[], unit=Unit("km")),
              phi=Angle(value=f32[], unit=Unit("deg")) )

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=current.r * jnp.sin(current.theta), phi=current.phi)


# =============================================================================
# MathSphericalPos


# -----------------------------------------------
# 1D


@dispatch
def represent_as(
    current: MathSphericalPos, target: type[CartesianPos1D], /, **kwargs: Any
) -> CartesianPos1D:
    """MathSphericalPos -> CartesianPos1D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.MathSphericalPos(r=u.Quantity(1.0, "km"), theta=u.Quantity(10.0, "deg"),
    ...                         phi=u.Quantity(14, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos1D)
    >>> x2
    CartesianPos1D(x=Quantity[...](value=f32[], unit=Unit("km")))

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.r * jnp.sin(current.phi) * jnp.cos(current.theta))


@dispatch
def represent_as(
    current: MathSphericalPos, target: type[RadialPos], /, **kwargs: Any
) -> RadialPos:
    """MathSphericalPos -> RadialPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.MathSphericalPos(r=u.Quantity(1.0, "km"), theta=u.Quantity(10.0, "deg"),
    ...                         phi=u.Quantity(14, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.RadialPos)
    >>> x2
    RadialPos(r=Distance(value=f32[], unit=Unit("km")))

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=current.r)


# -----------------------------------------------
# 2D


@dispatch
def represent_as(
    current: MathSphericalPos, target: type[CartesianPos2D], /, **kwargs: Any
) -> CartesianPos2D:
    """MathSphericalPos -> CartesianPos2D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.MathSphericalPos(r=u.Quantity(1.0, "km"), theta=u.Quantity(10.0, "deg"),
    ...                         phi=u.Quantity(14, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos2D)
    >>> x2
    CartesianPos2D( x=Quantity[...](value=f32[], unit=Unit("km")),
                    y=Quantity[...](value=f32[], unit=Unit("km")) )

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    x = current.r * jnp.sin(current.phi) * jnp.cos(current.theta)
    y = current.r * jnp.sin(current.phi) * jnp.sin(current.theta)
    return target(x=x, y=y)


@dispatch
def represent_as(
    current: MathSphericalPos, target: type[PolarPos], /, **kwargs: Any
) -> PolarPos:
    """MathSphericalPos -> PolarPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.MathSphericalPos(r=u.Quantity(1.0, "km"), theta=u.Quantity(10.0, "deg"),
    ...                         phi=u.Quantity(14, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.PolarPos)
    >>> x2
    PolarPos( r=Distance(value=f32[], unit=Unit("km")),
              phi=Angle(value=f32[], unit=Unit("deg")) )

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=current.r * jnp.sin(current.phi), phi=current.theta)


# =============================================================================
# ProlateSpheroidalPos


# -----------------------------------------------
# 1D


@dispatch
def represent_as(
    current: ProlateSpheroidalPos, target: type[AbstractPos1D], /, **kwargs: Any
) -> AbstractPos1D:
    """ProlateSpheroidalPos -> AbstractPos1D.

    Examples
    --------
    >>> import warnings
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.ProlateSpheroidalPos(
    ...     mu=Quantity(2.0, "kpc2"),
    ...     nu=Quantity(0.5, "kpc2"),
    ...     phi=Quantity(0.5, "rad"),
    ...     Delta=Quantity(1.0, "kpc"),
    ... )

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos1D)
    >>> x2
    CartesianPos1D(
      x=Quantity[PhysicalType('length')](value=f32[], unit=Unit("kpc"))
    )

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.RadialPos)
    >>> x2
    RadialPos(r=Distance(value=f32[], unit=Unit("kpc")))

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return represent_as(represent_as(current, CartesianPos3D), target)


# -----------------------------------------------
# 2D


@dispatch
def represent_as(
    current: ProlateSpheroidalPos, target: type[AbstractPos2D], /, **kwargs: Any
) -> AbstractPos2D:
    """ProlateSpheroidalPos -> AbstractPos2D.

    Examples
    --------
    >>> import warnings
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.ProlateSpheroidalPos(
    ...     mu=Quantity(2.0, "kpc2"),
    ...     nu=Quantity(0.5, "kpc2"),
    ...     phi=Quantity(0.5, "rad"),
    ...     Delta=Quantity(1.0, "kpc"),
    ... )

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos2D)
    >>> x2
    CartesianPos2D( x=Quantity[...](value=f32[], unit=Unit("kpc")),
                       y=Quantity[...](value=f32[], unit=Unit("kpc")) )

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.PolarPos)
    >>> x2
    PolarPos( r=Distance(value=f32[], unit=Unit("kpc")),
              phi=Angle(value=f32[], unit=Unit("rad")) )

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return represent_as(represent_as(current, CartesianPos3D), target)



"""Base classes for operators on coordinates and potentials."""

__all__ = ["IdentityOperator"]

from typing import Any, Literal, final

from .base import AbstractOperator


@final
class IdentityOperator(AbstractOperator):
    """Identity operation.

    This is the identity operation, which does nothing to the input.

    Examples
    --------
    We will work through many of the registered call signatures for the
    `IdentityOperator` class. Note that more call signatures may be registered.

    First, we make an instance of the operator:

    >>> import unxt as u
    >>> import coordinax as cx

    >>> op = cx.operators.IdentityOperator()
    >>> op
    IdentityOperator()

    And the common objects we will use:

    >>> q = u.Quantity([1, 2, 3], "kpc")
    >>> vec = cx.CartesianPos3D.from_(q)

    The first call signature is for the case where the input is a vector:

    >>> op(vec) is vec
    True

    The second call signature is for a Quantity:

    >>> op(q) is q
    True

    Actually, the `Identity` operator works for any vector or quantity:

    - 1D:

    >>> q = u.Quantity([1], "kpc")
    >>> vec = cx.CartesianPos1D.from_(q)
    >>> op(vec) is vec and op(q) is q
    True

    - 2D:

    >>> q = u.Quantity([1, 2], "kpc")
    >>> vec = cx.CartesianPos2D.from_(q)
    >>> op(vec) is vec and op(q) is q
    True

    - 3D (not using a `~coordinax.CartesianPos3D` instance):

    >>> q = u.Quantity([1, 2, 3], "kpc")
    >>> vec = cx.CartesianPos3D.from_(q).represent_as(cx.SphericalPos)
    >>> op(vec) is vec and op(q) is q
    True

    - 4D:

    >>> q = u.Quantity([1, 2, 3, 4], "kpc")  # 0th elt is ct
    >>> vec4 = cx.FourVector.from_(q)
    >>> op(vec4) is vec4 and op(q) is q
    True

    Lastly, many operators are time dependent and support a time argument. The`
    `Identity` operator will also pass through the time argument:

    >>> t = u.Quantity(0, "Gyr")
    >>> op(vec, t) == (vec, t)
    True

    >>> q = u.Quantity([1, 2, 3], "kpc")
    >>> op(q, t) == (q, t)
    True

    """

    @property
    def is_inertial(self) -> Literal[True]:
        """Identity operation is an inertial-frame preserving transform.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax.operators as cxo

        >>> op = cxo.IdentityOperator()
        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "IdentityOperator":
        """The inverse of the operator.

        Examples
        --------
        >>> import coordinax.operators as cxo

        >>> op = cxo.IdentityOperator()
        >>> op.inverse
        IdentityOperator()
        >>> op.inverse is op
        True

        """
        return self

    # -------------------------------------------
    # Dispatched call signatures
    # More call signatures are registered in the `coordinax._d<X>.operate` modules.

    @AbstractOperator.__call__.dispatch(precedence=1)
    def __call__(self: "IdentityOperator", arg: Any) -> Any:
        """Apply the Identity operation.

        This is the identity operation, which does nothing to the input.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> op = cx.operators.IdentityOperator()

        >>> q = u.Quantity([1, 2, 3], "kpc")
        >>> op(q) is q
        True

        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
        >>> op(vec) is vec
        True

        """
        return arg

    @AbstractOperator.__call__.dispatch(precedence=1)
    def __call__(self: "IdentityOperator", *args: Any) -> tuple[Any, ...]:
        """Apply the Identity operation.

        This is the identity operation, which does nothing to the input.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> op = cx.operators.IdentityOperator()

        >>> q = u.Quantity([1, 2, 3], "kpc")
        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
        >>> t = u.Quantity(10, "Gyr")

        >>> op(q, t) == (q, t)
        True

        >>> op(vec, t) == (vec, t)
        True

        """
        return args



"""Base classes for angular quantities."""

__all__: list[str] = []


import unxt as u
from unxt.quantity import AbstractQuantity

angle_dimension = u.dimension("angle")


class AbstractAngle(AbstractQuantity):  # type: ignore[misc]
    """Angular Quantity.

    See Also
    --------
    `coordinax.angle.Angle` : a concrete implementation of this class.

    Examples
    --------
    For this example, we will use the concrete implementation of
    `coordinax.angle.AbstractAngle`, `coordinax.angle.Angle`.

    >>> from coordinax.angle import Angle

    >>> a = Angle(90, "deg")
    >>> a
    Angle(Array(90, dtype=int32, weak_type=True), unit='deg')

    Angles have to have dimensions of angle.

    >>> try: Angle(90, "m")
    ... except ValueError as e: print(e)
    Angle must have dimensions angle.

    """

    def __check_init__(self) -> None:
        """Check the initialization."""
        if u.dimension_of(self) != angle_dimension:
            msg = "Angle must have dimensions angle."
            raise ValueError(msg)



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from typing import TypeAlias

from astropy.units import (
    CompositeUnit as AstropyCompositeUnit,
    Unit as AstropyUnit,
    UnitBase as AstropyUnitBase,
)
from jaxtyping import Float, Int, Shaped

from unxt.quantity import AbstractQuantity, Quantity

Unit: TypeAlias = AstropyUnit | AstropyUnitBase | AstropyCompositeUnit

FloatScalarQ = Float[AbstractQuantity, ""]
BatchFloatScalarQ = Shaped[FloatScalarQ, "*batch"]
BatchableFloatScalarQ = Shaped[FloatScalarQ, "*#batch"]

ScalarTime = Float[Quantity["time"], ""] | Int[Quantity["time"], ""]

BatchableTime = Shaped[Quantity["time"], "*#batch"]

BatchableArea = Shaped[Quantity["area"], "*#batch"]
BatchableDiffusivity = Shaped[Quantity["diffusivity"], "*#batch"]
BatchableSpecificEnergy = Shaped[Quantity["specific energy"], "*#batch"]

BatchableSpeed = Shaped[Quantity["speed"], "*#batch"]
BatchableAngularSpeed = Shaped[Quantity["angular speed"], "*#batch"]

BatchableAcc = Shaped[Quantity["acceleration"], "*#batch"]
BatchableAngularAcc = Shaped[Quantity["angular acceleration"], "*#batch"]


TimeBatchOrScalar = ScalarTime | BatchableTime



"""Type hints for `coordinax.distance`."""

__all__: list[str] = []


from jaxtyping import Shaped

import unxt as u

from .base import AbstractDistance

BatchableLength = Shaped[u.Quantity["length"], "*#batch"]
BatchableDistance = Shaped[AbstractDistance, "*#batch"]



"""Representation of coordinates in different systems."""

__all__ = ["AbstractAcc3D", "AbstractPos3D", "AbstractVel3D"]


from abc import abstractmethod
from typing_extensions import override

from coordinax._src.utils import classproperty
from coordinax._src.vectors.base import (
    AbstractAcc,
    AbstractPos,
    AbstractVector,
    AbstractVel,
)


class AbstractPos3D(AbstractPos):
    """Abstract representation of 3D coordinates in different systems."""

    @override
    @classproperty
    @classmethod
    def _cartesian_cls(cls) -> type[AbstractVector]:
        from .cartesian import CartesianPos3D

        return CartesianPos3D

    @classproperty
    @classmethod
    @abstractmethod
    def differential_cls(cls) -> type["AbstractVel3D"]:
        raise NotImplementedError


class AbstractVel3D(AbstractVel):
    """Abstract representation of 3D vector differentials."""

    @classproperty
    @classmethod
    def _cartesian_cls(cls) -> type[AbstractVector]:
        from .cartesian import CartesianVel3D

        return CartesianVel3D

    @classproperty
    @classmethod
    @abstractmethod
    def integral_cls(cls) -> type[AbstractPos3D]:
        raise NotImplementedError

    @classproperty
    @classmethod
    @abstractmethod
    def differential_cls(cls) -> type[AbstractAcc]:
        raise NotImplementedError


class AbstractAcc3D(AbstractAcc):
    """Abstract representation of 3D vector accelerations."""

    @classproperty
    @classmethod
    def _cartesian_cls(cls) -> type[AbstractVector]:
        from .cartesian import CartesianAcc3D

        return CartesianAcc3D

    @classproperty
    @classmethod
    @abstractmethod
    def integral_cls(cls) -> type[AbstractVel3D]:
        raise NotImplementedError



"""3-dimensional constructors."""
# pylint: disable=duplicate-code

__all__: list[str] = []

from typing import Any

from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D
from .cartesian import CartesianAcc3D, CartesianPos3D, CartesianVel3D

#####################################################################


@AbstractPos3D.from_.dispatch(precedence=-1)
def from_(cls: type[AbstractPos3D], obj: Any, /) -> CartesianPos3D:
    """Try to construct a 3D Cartesian position from an object.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = u.Quantity([1, 2, 3], "km")
    >>> cx.AbstractPos3D.from_(x)
    CartesianPos3D(
      x=Quantity[...](value=f32[], unit=Unit("km")),
      y=Quantity[...](value=f32[], unit=Unit("km")),
      z=Quantity[...](value=f32[], unit=Unit("km"))
    )

    """
    return obj if isinstance(obj, CartesianPos3D) else CartesianPos3D.from_(obj)


@AbstractPos3D.from_.dispatch(precedence=1)
def from_(cls: type[AbstractPos3D], obj: AbstractPos3D, /) -> AbstractPos3D:
    """Construct from a 3D position.

    Examples
    --------
    >>> import coordinax as cx

    >>> cart = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> cx.AbstractPos3D.from_(cart) is cart
    True

    >>> sph = cart.represent_as(cx.SphericalPos)
    >>> cx.AbstractPos3D.from_(sph) is sph
    True

    >>> cyl = cart.represent_as(cx.CylindricalPos)
    >>> cx.AbstractPos3D.from_(cyl) is cyl
    True

    """
    return obj


#####################################################################


@AbstractVel3D.from_.dispatch(precedence=-1)
def from_(cls: type[AbstractVel3D], obj: Any, /) -> CartesianVel3D:
    """Try to construct a 3D Cartesian velocity from an object.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = u.Quantity([1, 2, 3], "km / s")
    >>> cx.AbstractVel3D.from_(x)
    CartesianVel3D(
      d_x=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_y=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_z=Quantity[...]( value=f32[], unit=Unit("km / s") )
    )

    """
    return obj if isinstance(obj, CartesianVel3D) else CartesianVel3D.from_(obj)


@AbstractVel3D.from_.dispatch(precedence=1)
def from_(cls: type[AbstractVel3D], obj: AbstractVel3D, /) -> AbstractVel3D:
    """Construct from a 3D velocity.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.CartesianPos3D.from_([1, 1, 1], "km")

    >>> cart = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> cx.AbstractVel3D.from_(cart) is cart
    True

    >>> sph = cart.represent_as(cx.SphericalVel, q)
    >>> cx.AbstractVel3D.from_(sph) is sph
    True

    >>> cyl = cart.represent_as(cx.CylindricalVel, q)
    >>> cx.AbstractVel3D.from_(cyl) is cyl
    True

    """
    return obj


#####################################################################


@AbstractAcc3D.from_.dispatch(precedence=-1)
def from_(cls: type[AbstractAcc3D], obj: Any, /) -> CartesianAcc3D:
    """Try to construct a 3D Cartesian velocity from an object.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = u.Quantity([1, 2, 3], "km / s2")
    >>> cx.AbstractAcc3D.from_(x)
    CartesianAcc3D(
      d2_x=Quantity[...](value=f32[], unit=Unit("km / s2")),
      d2_y=Quantity[...](value=f32[], unit=Unit("km / s2")),
      d2_z=Quantity[...](value=f32[], unit=Unit("km / s2"))
    )

    """
    return obj if isinstance(obj, CartesianAcc3D) else CartesianAcc3D.from_(obj)


@AbstractAcc3D.from_.dispatch(precedence=1)
def from_(cls: type[AbstractAcc3D], obj: AbstractAcc3D, /) -> AbstractAcc3D:
    """Construct from a 3D velocity.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.CartesianPos3D.from_([1, 1, 1], "km")
    >>> p = cx.CartesianVel3D.from_([1, 1, 1], "km/s")

    >>> cart = cx.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> cx.AbstractAcc3D.from_(cart) is cart
    True

    >>> sph = cart.represent_as(cx.SphericalAcc, p, q)
    >>> cx.AbstractAcc3D.from_(sph) is sph
    True

    >>> cyl = cart.represent_as(cx.CylindricalAcc, p, q)
    >>> cx.AbstractAcc3D.from_(cyl) is cyl
    True

    """
    return obj



"""Coordinax interop with Astropy package."""

__all__: list[str] = []

import astropy.coordinates as apyc
import equinox as eqx

from coordinax._coordinax_space_frames import ICRS, Galactocentric
from coordinax._src.vectors.d3 import LonLatSphericalPos


@ICRS.from_._f.register  # noqa: SLF001
def from_(cls: type[ICRS], obj: apyc.ICRS, /) -> ICRS:
    """Construct from a `astropy.coordinates.ICRS`.

    Examples
    --------
    >>> import astropy.coordinates as apyc
    >>> import coordinax.frames as cxf

    >>> apy_icrs = apyc.ICRS()
    >>> cxf.ICRS.from_(apy_icrs)
    ICRS()

    """
    obj = eqx.error_if(obj, obj.has_data, "Astropy frame must not have data.")
    return cls()


@Galactocentric.from_._f.register  # noqa: SLF001
def from_(cls: type[Galactocentric], obj: apyc.Galactocentric, /) -> Galactocentric:
    """Construct from a `astropy.coordinates.Galactocentric`.

    Examples
    --------
    >>> import astropy.coordinates as apyc
    >>> import coordinax.frames as cxf

    >>> apy_gcf = apyc.Galactocentric()
    >>> apy_gcf
    <Galactocentric Frame (galcen_coord=<ICRS Coordinate: (ra, dec) in deg
    (266.4051, -28.936175)>, galcen_distance=8.122 kpc, galcen_v_sun=(12.9, 245.6, 7.78) km / s, z_sun=20.8 pc, roll=0.0 deg)>

    >>> gcf = cxf.Galactocentric.from_(apy_gcf)
    >>> gcf
    Galactocentric(
        galcen=LonLatSphericalPos( ... ),
        roll=Quantity[...](value=f32[], unit=Unit("deg")),
        z_sun=Quantity[...](value=f32[], unit=Unit("pc")),
        galcen_v_sun=CartesianVel3D( ... )
    )

    Checking equality

    >>> (gcf.galcen.lon.to_value("deg") == apy_gcf.galcen_coord.ra.to_value("deg")
    ...  and gcf.galcen.lat.to_value("deg") == apy_gcf.galcen_coord.dec.to_value("deg")
    ...  and gcf.galcen.distance.to_value("kpc") == apy_gcf.galcen_distance.to_value("kpc") )
    Array(True, dtype=bool)


    """  # noqa: E501
    obj = eqx.error_if(obj, obj.has_data, "Astropy frame must not have data.")
    galcen = LonLatSphericalPos(
        lon=obj.galcen_coord.ra,
        lat=obj.galcen_coord.dec,
        distance=obj.galcen_distance,
    )
    return cls(galcen, roll=obj.roll, z_sun=obj.z_sun, galcen_v_sun=obj.galcen_v_sun)



"""Base classes for operators on coordinates and potentials."""

__all__ = ["AbstractOperator"]

from abc import abstractmethod
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

import equinox as eqx
from jaxtyping import ArrayLike
from plum import dispatch

import unxt as u
from dataclassish import field_items

from coordinax._src.vectors.base import AbstractPos

if TYPE_CHECKING:
    from coordinax.operators import OperatorSequence


class AbstractOperator(eqx.Module):  # type: ignore[misc]
    """Abstract base class for operators on coordinates and potentials.

    An operator is an object that defines a transformation on coordinates. It
    can be applied to a set of coordinates to produce a new set of coordinates.
    Operators can be composed together to form a sequence of transformations.

    When defining a new operator, it should be able to work on
    `coordinax.AbstractVector` objects. If it is a spatial operator, it should
    also be able to work on (`coordinax.AbstractPos`, `unxt.Quantity['time']`)
    pairs (and then also `coordinax.FourVector` objects). If the vector can be
    created from a `unxt.Quantity` object, then the operator should also be able
    to work on `unxt.Quantity` object.

    """

    # ---------------------------------------------------------------
    # Constructors

    @classmethod
    @dispatch(precedence=-1)
    def from_(
        cls: "type[AbstractOperator]", *args: object, **kwargs: object
    ) -> "AbstractOperator":
        """Construct from a set of arguments.

        This is a low-priority dispatch that will be called if no other
        dispatch is found. It just tries to pass the arguments to the
        constructor.

        """
        return cls(*args, **kwargs)

    @classmethod
    @dispatch
    def from_(
        cls: "type[AbstractOperator]", obj: Mapping[str, Any], /
    ) -> "AbstractOperator":
        """Construct from a mapping.

        Examples
        --------
        >>> import coordinax.operators as co
        >>> operators = co.IdentityOperator() | co.IdentityOperator()
        >>> co.OperatorSequence.from_({"operators": operators})
        OperatorSequence(operators=(IdentityOperator(), IdentityOperator()))

        """
        return cls(**obj)

    @classmethod
    @dispatch
    def from_(
        cls: "type[AbstractOperator]",
        x: ArrayLike | list[float | int],
        unit: str,  # TODO: support unit object
        /,
    ) -> "AbstractOperator":
        """Construct from a Quantity's value and unit.

        Examples
        --------
        >>> import coordinax.operators as cxo

        >>> op = cxo.GalileanSpatialTranslationOperator.from_([1, 1, 1], "kpc")
        >>> print(op.translation)
        <CartesianPos3D (x[kpc], y[kpc], z[kpc])
            [1. 1. 1.]>

        >>> op = cxo.GalileanTranslationOperator.from_([3e5, 1, 1, 1], "kpc")
        >>> print(op.translation)
        <FourVector (t[kpc s / km], q=(x[kpc], y[kpc], z[kpc]))
            [1.001 1.    1.    1.   ]>

        >>> op = cxo.GalileanBoostOperator.from_([1, 1, 1], "km/s")
        >>> print(op.velocity)
        <CartesianVel3D (d_x[km / s], d_y[km / s], d_z[km / s])
            [1. 1. 1.]>

        """
        return cls(u.Quantity(x, unit))

    # -------------------------------------------

    @dispatch.abstract
    def __call__(
        self: "AbstractOperator",
        x: AbstractPos,  # noqa: ARG002
        /,
    ) -> AbstractPos:
        """Apply the operator to the coordinates `x`."""
        msg = "implement this method in the subclass"
        raise TypeError(msg)

    @dispatch.abstract
    def __call__(
        self: "AbstractOperator",
        x: AbstractPos,  # noqa: ARG002
        t: u.Quantity["time"],  # noqa: ARG002
        /,
    ) -> AbstractPos:
        """Apply the operator to the coordinates `x` at a time `t`."""
        msg = "implement this method in the subclass"
        raise TypeError(msg)

    # -------------------------------------------

    @property
    @abstractmethod
    def is_inertial(self) -> bool:
        """Whether the operation maintains an inertial reference frame."""
        ...

    @property
    @abstractmethod
    def inverse(self) -> "AbstractOperator":
        """The inverse of the operator."""
        ...

    # ===========================================
    # Sequence

    def __or__(self, other: "AbstractOperator") -> "OperatorSequence":
        """Compose with another operator."""
        from .sequential import OperatorSequence

        if isinstance(other, OperatorSequence):
            return other.__ror__(self)
        return OperatorSequence((self, other))


# TODO: move to the class in py3.11+
@AbstractOperator.from_.dispatch  # type: ignore[attr-defined, misc]
def from_(cls: type[AbstractOperator], obj: AbstractOperator, /) -> AbstractOperator:
    """Construct an operator from another operator.

    Examples
    --------
    >>> import coordinax.operators as cxo

    If the object is the same type, it should return the object itself.

    >>> op = cxo.IdentityOperator()
    >>> cxo.IdentityOperator.from_(op) is op
    True

    If the object is a different type, it will error.

    >>> try:
    ...     cxo.GalileanBoostOperator.from_(op)
    ... except TypeError as e:
    ...     print(e)
    Cannot construct <class 'coordinax...GalileanBoostOperator'> from <class 'coordinax...IdentityOperator'>.

    Unless the object is a subclass of the target class.

    >>> class MyOperator(cxo.IdentityOperator):
    ...     pass

    >>> op = MyOperator()
    >>> op
    MyOperator()

    >>> newop = cxo.IdentityOperator.from_(op)
    >>> newop is op, isinstance(newop, cxo.IdentityOperator)
    (False, True)

    """  # noqa: E501
    if not isinstance(obj, cls):
        msg = f"Cannot construct {cls} from {type(obj)}."
        raise TypeError(msg)

    # avoid copying if the types are the same. Isinstance is not strict
    # enough, so we use type() instead.
    if type(obj) is cls:  # pylint: disable=unidiomatic-typecheck
        return obj

    return cls(**dict(field_items(obj)))



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from typing import Any

from plum import dispatch

from .cartesian import CartesianAccND, CartesianPosND, CartesianVelND
from .poincare import PoincarePolarVector

###############################################################################
# Cartesian


@dispatch
def represent_as(
    current: CartesianPosND, target: type[CartesianPosND], /, **kwargs: Any
) -> CartesianPosND:
    """CartesianPosND -> CartesianPosND.

    Examples
    --------
    >>> import coordinax as cx
    >>> import unxt as u

    >>> q = cx.CartesianPosND(u.Quantity([1, 2, 3, 4], "kpc"))

    >>> cx.represent_as(q, cx.CartesianPosND) is q
    True

    """
    return current


@dispatch
def represent_as(
    current: CartesianVelND,
    target: type[CartesianVelND],
    position: CartesianPosND,
    /,
    **kwargs: Any,
) -> CartesianVelND:
    """CartesianVelND -> CartesianVelND.

    Examples
    --------
    >>> import coordinax as cx

    >>> x = cx.CartesianPosND.from_([1, 2, 3, 4], "km")
    >>> v = cx.CartesianVelND.from_([1, 2, 3, 4], "km/s")

    >>> cx.represent_as(v, cx.CartesianVelND, x) is v
    True

    """
    return current


# =============================================================================
# CartesianVelND


@dispatch
def represent_as(
    current: CartesianVelND, target: type[CartesianVelND], /
) -> CartesianVelND:
    """CartesianVelND -> CartesianVelND with no position.

    Cartesian coordinates are an affine coordinate system and so the
    transformation of an n-th order derivative vector in this system do not
    require lower-order derivatives to be specified. See
    https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for more
    information. This mixin provides a corresponding implementation of the
    `coordinax.represent_as` method for Cartesian velocities.

    Examples
    --------
    >>> import coordinax as cx
    >>> v = cx.CartesianVelND.from_([1, 1, 1], "m/s")
    >>> cx.represent_as(v, cx.CartesianVelND) is v
    True

    """
    return current


# =============================================================================
# CartesianAccND


@dispatch
def represent_as(
    current: CartesianAccND, target: type[CartesianAccND], /
) -> CartesianAccND:
    """CartesianAccND -> CartesianAccND with no position.

    Cartesian coordinates are an affine coordinate system and so the
    transformation of an n-th order derivative vector in this system do not
    require lower-order derivatives to be specified. See
    https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for more
    information. This mixin provides a corresponding implementation of the
    `coordinax.represent_as` method for Cartesian vectors.

    Examples
    --------
    >>> import coordinax as cx
    >>> a = cx.CartesianAccND.from_([1, 1, 1], "m/s2")
    >>> cx.represent_as(a, cx.CartesianAccND) is a
    True

    """
    return current


###############################################################################
# Poincare


@dispatch
def represent_as(
    current: PoincarePolarVector, target: type[PoincarePolarVector], /, **kwargs: Any
) -> PoincarePolarVector:
    """PoincarePolarVector -> PoincarePolarVector."""
    return current



# mypy: disable-error-code="attr-defined"

"""Interoperability with :mod:`astropy.coordinates`."""

__all__: list[str] = []


import astropy.coordinates as apyc
import astropy.units as apyu
from jaxtyping import Shaped
from plum import conversion_method, convert

import unxt as u

import coordinax as cx

#####################################################################

# =====================================
# Quantity


@conversion_method(cx.AbstractPos3D, apyu.Quantity)  # type: ignore[misc]
def vec_to_q(obj: cx.AbstractPos3D, /) -> Shaped[apyu.Quantity, "*batch 3"]:
    """`coordinax.AbstractPos3D` -> `astropy.units.Quantity`.

    Examples
    --------
    >>> import coordinax as cx
    >>> from plum import convert
    >>> import astropy.units as apyu

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> convert(vec, apyu.Quantity)
    <Quantity [1., 2., 3.] kpc>

    >>> vec = cx.SphericalPos(r=apyu.Quantity(1, unit="kpc"),
    ...                       theta=apyu.Quantity(2, unit="deg"),
    ...                       phi=apyu.Quantity(3, unit="deg"))
    >>> convert(vec, apyu.Quantity)
    <Quantity [0.03485167, 0.0018265 , 0.99939084] kpc>

    >>> vec = cx.CylindricalPos(rho=apyu.Quantity(1, unit="kpc"),
    ...                         phi=apyu.Quantity(2, unit="deg"),
    ...                         z=apyu.Quantity(3, unit="pc"))
    >>> convert(vec, apyu.Quantity)
    <Quantity [0.99939084, 0.0348995 , 0.003     ] kpc>

    """
    return convert(convert(obj, u.Quantity), apyu.Quantity)


@conversion_method(cx.CartesianAcc3D, apyu.Quantity)  # type: ignore[misc]
@conversion_method(cx.CartesianVel3D, apyu.Quantity)  # type: ignore[misc]
def vec_diff_to_q(
    obj: cx.CartesianVel3D | cx.CartesianAcc3D, /
) -> Shaped[apyu.Quantity, "*batch 3"]:
    """`coordinax.CartesianVel3D` -> `astropy.units.Quantity`.

    Examples
    --------
    >>> import coordinax as cx
    >>> from plum import convert
    >>> from astropy.units import Quantity as AstropyQuantity

    >>> dif = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> convert(dif, AstropyQuantity)
    <Quantity [1., 2., 3.] km / s>

    >>> dif2 = cx.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> convert(dif2, AstropyQuantity)
    <Quantity [1., 2., 3.] km / s2>

    """
    return convert(convert(obj, u.Quantity), apyu.Quantity)


# =====================================
# CartesianPos3D


@conversion_method(cx.CartesianPos3D, apyc.BaseRepresentation)  # type: ignore[misc]
@conversion_method(cx.CartesianPos3D, apyc.CartesianRepresentation)  # type: ignore[misc]
def cart3_to_apycart3(obj: cx.CartesianPos3D, /) -> apyc.CartesianRepresentation:
    """`coordinax.CartesianPos3D` -> `astropy.CartesianRepresentation`.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> convert(vec, apyc.CartesianRepresentation)
    <CartesianRepresentation (x, y, z) in kpc
        (1., 2., 3.)>

    >>> convert(vec, apyc.BaseRepresentation)
    <CartesianRepresentation (x, y, z) in kpc
        (1., 2., 3.)>

    """
    return apyc.CartesianRepresentation(
        x=convert(obj.x, apyu.Quantity),
        y=convert(obj.y, apyu.Quantity),
        z=convert(obj.z, apyu.Quantity),
    )


@conversion_method(apyc.CartesianRepresentation, cx.CartesianPos3D)  # type: ignore[misc]
def apycart3_to_cart3(obj: apyc.CartesianRepresentation, /) -> cx.CartesianPos3D:
    """`astropy.CartesianRepresentation` -> `coordinax.CartesianPos3D`.

    Examples
    --------
    >>> import coordinax as cx
    >>> from astropy.coordinates import CartesianRepresentation

    >>> vec = CartesianRepresentation(1, 2, 3, unit="kpc")
    >>> convert(vec, cx.CartesianPos3D)
    CartesianPos3D(
      x=Quantity[PhysicalType('length')](value=f32[], unit=Unit("kpc")),
      y=Quantity[PhysicalType('length')](value=f32[], unit=Unit("kpc")),
      z=Quantity[PhysicalType('length')](value=f32[], unit=Unit("kpc"))
    )

    """
    return cx.CartesianPos3D.from_(obj)


# =====================================
# CylindricalPos


@conversion_method(cx.CylindricalPos, apyc.BaseRepresentation)  # type: ignore[misc]
@conversion_method(cx.CylindricalPos, apyc.CylindricalRepresentation)  # type: ignore[misc]
def cyl_to_apycyl(obj: cx.CylindricalPos, /) -> apyc.CylindricalRepresentation:
    """`coordinax.CylindricalPos` -> `astropy.CylindricalRepresentation`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CylindricalPos(rho=u.Quantity(1, unit="kpc"),
    ...                         phi=u.Quantity(2, unit="deg"),
    ...                         z=u.Quantity(3, unit="pc"))
    >>> convert(vec, apyc.CylindricalRepresentation)
    <CylindricalRepresentation (rho, phi, z) in (kpc, deg, pc)
        (1., 2., 3.)>

    >>> convert(vec, apyc.BaseRepresentation)
    <CylindricalRepresentation (rho, phi, z) in (kpc, deg, pc)
        (1., 2., 3.)>

    """
    return apyc.CylindricalRepresentation(
        rho=convert(obj.rho, apyu.Quantity),
        phi=convert(obj.phi, apyu.Quantity),
        z=convert(obj.z, apyu.Quantity),
    )


@conversion_method(apyc.CylindricalRepresentation, cx.CylindricalPos)  # type: ignore[misc]
def apycyl_to_cyl(obj: apyc.CylindricalRepresentation, /) -> cx.CylindricalPos:
    """`astropy.CylindricalRepresentation` -> `coordinax.CylindricalPos`.

    Examples
    --------
    >>> import astropy.units as u
    >>> import coordinax as cx
    >>> from astropy.coordinates import CylindricalRepresentation

    >>> cyl = CylindricalRepresentation(rho=1 * u.kpc, phi=2 * u.deg, z=30 * u.pc)
    >>> convert(cyl, cx.CylindricalPos)
    CylindricalPos(
        rho=Quantity[...](value=f32[], unit=Unit("kpc")),
        phi=Angle(value=f32[], unit=Unit("deg")),
        z=Quantity[...](value=f32[], unit=Unit("pc"))
    )

    """
    return cx.CylindricalPos.from_(obj)


# =====================================
# SphericalPos


@conversion_method(cx.SphericalPos, apyc.BaseRepresentation)  # type: ignore[misc]
@conversion_method(cx.SphericalPos, apyc.PhysicsSphericalRepresentation)  # type: ignore[misc]
def sph_to_apysph(obj: cx.SphericalPos, /) -> apyc.PhysicsSphericalRepresentation:
    """`coordinax.SphericalPos` -> `astropy.PhysicsSphericalRepresentation`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.SphericalPos(r=u.Quantity(1, unit="kpc"),
    ...                       theta=u.Quantity(2, unit="deg"),
    ...                       phi=u.Quantity(3, unit="deg"))
    >>> convert(vec, apyc.PhysicsSphericalRepresentation)
    <PhysicsSphericalRepresentation (phi, theta, r) in (deg, deg, kpc)
        (3., 2., 1.)>

    """
    return apyc.PhysicsSphericalRepresentation(
        r=convert(obj.r, apyu.Quantity),
        phi=convert(obj.phi, apyu.Quantity),
        theta=convert(obj.theta, apyu.Quantity),
    )


@conversion_method(apyc.PhysicsSphericalRepresentation, cx.SphericalPos)  # type: ignore[misc]
def apysph_to_sph(obj: apyc.PhysicsSphericalRepresentation, /) -> cx.SphericalPos:
    """`astropy.PhysicsSphericalRepresentation` -> `coordinax.SphericalPos`.

    Examples
    --------
    >>> import astropy.units as u
    >>> import coordinax as cx
    >>> from astropy.coordinates import PhysicsSphericalRepresentation

    >>> sph = PhysicsSphericalRepresentation(r=1 * u.kpc, theta=2 * u.deg,
    ...                                      phi=3 * u.deg)
    >>> convert(sph, cx.SphericalPos)
    SphericalPos(
      r=Distance(value=f32[], unit=Unit("kpc")),
      theta=Angle(value=f32[], unit=Unit("deg")),
      phi=Angle(value=f32[], unit=Unit("deg"))
    )

    """
    return cx.SphericalPos.from_(obj)


# =====================================
# LonLatSphericalPos


@conversion_method(cx.LonLatSphericalPos, apyc.BaseRepresentation)  # type: ignore[misc]
@conversion_method(cx.LonLatSphericalPos, apyc.PhysicsSphericalRepresentation)  # type: ignore[misc]
def lonlatsph_to_apysph(obj: cx.LonLatSphericalPos, /) -> apyc.SphericalRepresentation:
    """`coordinax.LonLatSphericalPos` -> `astropy.SphericalRepresentation`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.LonLatSphericalPos(lon=u.Quantity(2, unit="deg"),
    ...                             lat=u.Quantity(3, unit="deg"),
    ...                             distance=u.Quantity(1, unit="kpc"))
    >>> convert(vec, apyc.SphericalRepresentation)
    <SphericalRepresentation (lon, lat, distance) in (deg, deg, kpc)
        (2., 3., 1.)>

    """
    return apyc.SphericalRepresentation(
        lon=convert(obj.lon, apyu.Quantity),
        lat=convert(obj.lat, apyu.Quantity),
        distance=convert(obj.distance, apyu.Quantity),
    )


@conversion_method(apyc.SphericalRepresentation, cx.LonLatSphericalPos)  # type: ignore[misc]
def apysph_to_lonlatsph(obj: apyc.SphericalRepresentation, /) -> cx.LonLatSphericalPos:
    """`astropy.SphericalRepresentation` -> `coordinax.LonLatSphericalPos`.

    Examples
    --------
    >>> import astropy.units as u
    >>> import coordinax as cx
    >>> from astropy.coordinates import SphericalRepresentation

    >>> sph = SphericalRepresentation(lon=2 * u.deg, lat=3 * u.deg,
    ...                               distance=1 * u.kpc)
    >>> convert(sph, cx.LonLatSphericalPos)
    LonLatSphericalPos(
      lon=Angle(value=f32[], unit=Unit("deg")),
      lat=Angle(value=f32[], unit=Unit("deg")),
      distance=Distance(value=f32[], unit=Unit("kpc"))
    )

    """
    return cx.LonLatSphericalPos.from_(obj)


# =====================================
# CartesianVel3D


@conversion_method(cx.CartesianVel3D, apyc.BaseDifferential)  # type: ignore[misc]
@conversion_method(cx.CartesianVel3D, apyc.CartesianDifferential)  # type: ignore[misc]
def diffcart3_to_apycart3(obj: cx.CartesianVel3D, /) -> apyc.CartesianDifferential:
    """`coordinax.CartesianVel3D` -> `astropy.CartesianDifferential`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> dif = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> convert(dif, apyc.CartesianDifferential)
    <CartesianDifferential (d_x, d_y, d_z) in km / s
        (1., 2., 3.)>

    """
    return apyc.CartesianDifferential(
        d_x=convert(obj.d_x, apyu.Quantity),
        d_y=convert(obj.d_y, apyu.Quantity),
        d_z=convert(obj.d_z, apyu.Quantity),
    )


@conversion_method(  # type: ignore[misc]
    apyc.CartesianDifferential, cx.CartesianVel3D
)
def apycart3_to_diffcart3(obj: apyc.CartesianDifferential, /) -> cx.CartesianVel3D:
    """`astropy.CartesianDifferential` -> `coordinax.CartesianVel3D`.

    Examples
    --------
    >>> import coordinax as cx
    >>> from astropy.coordinates import CartesianDifferential

    >>> dcart = CartesianDifferential(1, 2, 3, unit="km/s")
    >>> convert(dcart, cx.CartesianVel3D)
    CartesianVel3D(
      d_x=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_y=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_z=Quantity[...]( value=f32[], unit=Unit("km / s") )
    )

    """
    return cx.CartesianVel3D.from_(obj)


# =====================================
# CylindricalVel


@conversion_method(cx.CylindricalVel, apyc.BaseDifferential)  # type: ignore[misc]
@conversion_method(cx.CylindricalVel, apyc.CylindricalDifferential)  # type: ignore[misc]
def diffcyl_to_apycyl(obj: cx.CylindricalVel, /) -> apyc.CylindricalDifferential:
    """`coordinax.CylindricalVel` -> `astropy.CylindricalDifferential`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> import astropy.coordinates as apyc

    >>> dif = cx.CylindricalVel(d_rho=u.Quantity(1, unit="km/s"),
    ...                         d_phi=u.Quantity(2, unit="mas/yr"),
    ...                         d_z=u.Quantity(3, unit="km/s"))
    >>> convert(dif, apyc.CylindricalDifferential)
    <CylindricalDifferential (d_rho, d_phi, d_z) in (km / s, mas / yr, km / s)
        (1., 2., 3.)>

    >>> convert(dif, apyc.BaseDifferential)
    <CylindricalDifferential (d_rho, d_phi, d_z) in (km / s, mas / yr, km / s)
        (1., 2., 3.)>

    """
    return apyc.CylindricalDifferential(
        d_rho=convert(obj.d_rho, apyu.Quantity),
        d_phi=convert(obj.d_phi, apyu.Quantity),
        d_z=convert(obj.d_z, apyu.Quantity),
    )


@conversion_method(  # type: ignore[misc]
    apyc.CylindricalDifferential, cx.CylindricalVel
)
def apycyl_to_diffcyl(obj: apyc.CylindricalDifferential, /) -> cx.CylindricalVel:
    """`astropy.CylindricalVel` -> `coordinax.CylindricalVel`.

    Examples
    --------
    >>> import astropy.units as u
    >>> import astropy.coordinates as apyc
    >>> import coordinax as cx

    >>> dcyl = apyc.CylindricalDifferential(d_rho=1 * u.km / u.s, d_phi=2 * u.mas/u.yr,
    ...                                     d_z=2 * u.km / u.s)
    >>> convert(dcyl, cx.CylindricalVel)
    CylindricalVel(
      d_rho=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_phi=Quantity[...]( value=f32[], unit=Unit("mas / yr") ),
      d_z=Quantity[...]( value=f32[], unit=Unit("km / s") )
    )

    """
    return cx.CylindricalVel.from_(obj)


# =====================================
# SphericalVel


@conversion_method(cx.SphericalVel, apyc.BaseDifferential)  # type: ignore[misc]
@conversion_method(cx.SphericalVel, apyc.PhysicsSphericalDifferential)  # type: ignore[misc]
def diffsph_to_apysph(obj: cx.SphericalVel, /) -> apyc.PhysicsSphericalDifferential:
    """SphericalVel -> `astropy.PhysicsSphericalDifferential`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> dif = cx.SphericalVel(d_r=u.Quantity(1, unit="km/s"),
    ...                       d_theta=u.Quantity(2, unit="mas/yr"),
    ...                       d_phi=u.Quantity(3, unit="mas/yr"))
    >>> convert(dif, apyc.PhysicsSphericalDifferential)
    <PhysicsSphericalDifferential (d_phi, d_theta, d_r) in (mas / yr, mas / yr, km / s)
        (3., 2., 1.)>

    >>> convert(dif, apyc.BaseDifferential)
    <PhysicsSphericalDifferential (d_phi, d_theta, d_r) in (mas / yr, mas / yr, km / s)
        (3., 2., 1.)>

    """
    return apyc.PhysicsSphericalDifferential(
        d_r=convert(obj.d_r, apyu.Quantity),
        d_theta=convert(obj.d_theta, apyu.Quantity),
        d_phi=convert(obj.d_phi, apyu.Quantity),
    )


@conversion_method(  # type: ignore[misc]
    apyc.PhysicsSphericalDifferential, cx.SphericalVel
)
def apysph_to_diffsph(obj: apyc.PhysicsSphericalDifferential, /) -> cx.SphericalVel:
    """`astropy.PhysicsSphericalDifferential` -> SphericalVel.

    Examples
    --------
    >>> import astropy.units as u
    >>> import coordinax as cx
    >>> from astropy.coordinates import PhysicsSphericalDifferential

    >>> dif = PhysicsSphericalDifferential(d_r=1 * u.km / u.s, d_theta=2 * u.mas/u.yr,
    ...                                    d_phi=3 * u.mas/u.yr)
    >>> convert(dif, cx.SphericalVel)
    SphericalVel(
      d_r=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_theta=Quantity[...]( value=f32[], unit=Unit("mas / yr") ),
      d_phi=Quantity[...]( value=f32[], unit=Unit("mas / yr") )
    )

    """
    return cx.SphericalVel.from_(obj)


# =====================================
# LonLatSphericalVel


@conversion_method(cx.LonLatSphericalVel, apyc.BaseDifferential)  # type: ignore[misc]
@conversion_method(cx.LonLatSphericalVel, apyc.SphericalDifferential)  # type: ignore[misc]
def difflonlatsph_to_apysph(
    obj: cx.LonLatSphericalVel, /
) -> apyc.SphericalDifferential:
    """LonLatSphericalVel -> `astropy.SphericalVel`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> dif = cx.LonLatSphericalVel(d_distance=u.Quantity(1, unit="km/s"),
    ...                             d_lat=u.Quantity(2, unit="mas/yr"),
    ...                             d_lon=u.Quantity(3, unit="mas/yr"))
    >>> convert(dif, apyc.SphericalDifferential)
    <SphericalDifferential (d_lon, d_lat, d_distance) in (mas / yr, mas / yr, km / s)
        (3., 2., 1.)>

    >>> convert(dif, apyc.BaseDifferential)
    <SphericalDifferential (d_lon, d_lat, d_distance) in (mas / yr, mas / yr, km / s)
        (3., 2., 1.)>

    """
    return apyc.SphericalDifferential(
        d_distance=convert(obj.d_distance, apyu.Quantity),
        d_lon=convert(obj.d_lon, apyu.Quantity),
        d_lat=convert(obj.d_lat, apyu.Quantity),
    )


@conversion_method(  # type: ignore[misc]
    apyc.SphericalDifferential, cx.LonLatSphericalVel
)
def apysph_to_difflonlatsph(
    obj: apyc.SphericalDifferential, /
) -> cx.LonLatSphericalVel:
    """`astropy.coordinates.SphericalDifferential` -> LonLatSphericalVel.

    Examples
    --------
    >>> import astropy.units as u
    >>> import coordinax as cx
    >>> from astropy.coordinates import SphericalDifferential

    >>> dif = SphericalDifferential(d_distance=1 * u.km / u.s, d_lat=2 * u.mas/u.yr,
    ...                             d_lon=3 * u.mas/u.yr)
    >>> convert(dif, cx.LonLatSphericalVel)
    LonLatSphericalVel(
      d_lon=Quantity[...]( value=f32[], unit=Unit("mas / yr") ),
      d_lat=Quantity[...]( value=f32[], unit=Unit("mas / yr") ),
      d_distance=Quantity[...]( value=f32[], unit=Unit("km / s") )
    )

    """
    return cx.LonLatSphericalVel.from_(obj)


# =====================================
# LonCosLatSphericalVel


@conversion_method(cx.LonCosLatSphericalVel, apyc.BaseDifferential)  # type: ignore[misc]
@conversion_method(cx.LonCosLatSphericalVel, apyc.SphericalCosLatDifferential)  # type: ignore[misc]
def diffloncoslatsph_to_apysph(
    obj: cx.LonCosLatSphericalVel, /
) -> apyc.SphericalCosLatDifferential:
    """LonCosLatSphericalVel -> `astropy.SphericalCosLatDifferential`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> dif = cx.LonCosLatSphericalVel(d_distance=u.Quantity(1, unit="km/s"),
    ...                                d_lat=u.Quantity(2, unit="mas/yr"),
    ...                                d_lon_coslat=u.Quantity(3, unit="mas/yr"))
    >>> convert(dif, apyc.SphericalCosLatDifferential)
    <SphericalCosLatDifferential (d_lon_coslat, d_lat, d_distance) in (mas / yr, mas / yr, km / s)
        (3., 2., 1.)>

    >>> convert(dif, apyc.BaseDifferential)
    <SphericalCosLatDifferential (d_lon_coslat, d_lat, d_distance) in (mas / yr, mas / yr, km / s)
        (3., 2., 1.)>

    """  # noqa: E501
    return apyc.SphericalCosLatDifferential(
        d_distance=convert(obj.d_distance, apyu.Quantity),
        d_lon_coslat=convert(obj.d_lon_coslat, apyu.Quantity),
        d_lat=convert(obj.d_lat, apyu.Quantity),
    )


@conversion_method(  # type: ignore[misc]
    apyc.SphericalCosLatDifferential, cx.LonCosLatSphericalVel
)
def apysph_to_diffloncoslatsph(
    obj: apyc.SphericalCosLatDifferential, /
) -> cx.LonCosLatSphericalVel:
    """`astropy.SphericalCosLatDifferential` -> LonCosLatSphericalVel.

    Examples
    --------
    >>> import astropy.units as u
    >>> import coordinax as cx
    >>> from astropy.coordinates import SphericalCosLatDifferential

    >>> dif = SphericalCosLatDifferential(d_distance=1 * u.km / u.s,
    ...                                   d_lat=2 * u.mas/u.yr,
    ...                                   d_lon_coslat=3 * u.mas/u.yr)
    >>> convert(dif, cx.LonCosLatSphericalVel)
    LonCosLatSphericalVel(
      d_lon_coslat=Quantity[...]( value=f32[], unit=Unit("mas / yr") ),
      d_lat=Quantity[...]( value=f32[], unit=Unit("mas / yr") ),
      d_distance=Quantity[...]( value=f32[], unit=Unit("km / s") )
    )

    """
    return cx.LonCosLatSphericalVel.from_(obj)



# ruff: noqa: ERA001
"""Galilean coordinate transformations."""

__all__ = ["GalileanBoostOperator"]


from dataclasses import replace
from typing import Any, Literal, final

import equinox as eqx
import jax.numpy as jnp
from plum import convert, dispatch

import quaxed.numpy as jnp
import unxt as u

from .base import AbstractGalileanOperator
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.operators.identity import IdentityOperator
from coordinax._src.vectors.d3 import AbstractPos3D, CartesianVel3D
from coordinax._src.vectors.d4 import FourVector


@final
class GalileanBoostOperator(AbstractGalileanOperator):
    r"""Operator for Galilean boosts.

    The coordinate transform is given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t, \mathbf{x} + \mathbf{v} t)

    where :math:`\mathbf{v}` is the boost velocity.

    Parameters
    ----------
    velocity : :class:`vector.CartesianVel3D`
        The boost velocity. This parameters uses
        :meth:`vector.CartesianVel3D.from_` to enable a variety of more
        convenient input types. See :class:`vector.CartesianVel3D` for details.

    Examples
    --------
    >>> import coordinax.operators as cxo

    >>> op = cxo.GalileanBoostOperator.from_([1.0, 2.0, 3.0], "m/s")
    >>> op
    GalileanBoostOperator( velocity=CartesianVel3D( ... ) )

    Note that the velocity is a :class:`vector.CartesianVel3D`, which was
    constructed from an array, using :meth:`vector.CartesianVel3D.from_`.

    """

    velocity: CartesianVel3D = eqx.field(converter=CartesianVel3D.from_)
    """The boost velocity.

    This parameters uses :meth:`vector.CartesianVel3D.from_` to enable a variety
    of more convenient input types. See :class:`vector.CartesianVel3D` for
    details.
    """

    # -----------------------------------------------------

    @property
    def is_inertial(self) -> Literal[True]:
        """Galilean boost is an inertial-frame preserving transform.

        Examples
        --------
        >>> import coordinax.operators as cxo

        >>> op = cxo.GalileanBoostOperator.from_([1, 2, 3], "m/s")
        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "GalileanBoostOperator":
        """The inverse of the operator.

        Examples
        --------
        >>> import coordinax.operators as cxo

        >>> op = cxo.GalileanBoostOperator.from_([1, 2, 3], "m/s")
        >>> op.inverse
        GalileanBoostOperator( velocity=CartesianVel3D( ... ) )

        >>> print(op.inverse.velocity)
        <CartesianVel3D (d_x[m / s], d_y[m / s], d_z[m / s])
            [-1. -2. -3.]>

        """
        return GalileanBoostOperator(-self.velocity)

    # -----------------------------------------------------

    @AbstractOperator.__call__.dispatch(precedence=1)
    def __call__(
        self: "GalileanBoostOperator", q: AbstractPos3D, t: u.Quantity["time"], /
    ) -> tuple[AbstractPos3D, u.Quantity["time"]]:
        """Apply the boost to the coordinates.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> op = cx.operators.GalileanBoostOperator.from_([1, 2, 3], "m/s")

        >>> q = cx.CartesianPos3D.from_([0, 0, 0], "m")
        >>> t = u.Quantity(1, "s")
        >>> newq, newt = op(q, t)
        >>> newt
        Quantity['time'](Array(1, dtype=int32, ...), unit='s')
        >>> print(newq)
        <CartesianPos3D (x[m], y[m], z[m])
            [1. 2. 3.]>

        """
        return q + self.velocity * t, t

    @AbstractOperator.__call__.dispatch
    def __call__(self: "GalileanBoostOperator", v4: FourVector, /) -> FourVector:
        """Apply the boost to the coordinates."""  # TODO: add example
        return replace(v4, q=v4.q + self.velocity * v4.t)


@dispatch  # type: ignore[misc]
def simplify_op(op: GalileanBoostOperator, /, **kwargs: Any) -> AbstractOperator:
    """Simplify a boost operator.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.operators as cxo

    An operator with real effect cannot be simplified:

    >>> op = cxo.GalileanBoostOperator.from_([1, 0, 0], "m/s")
    >>> cxo.simplify_op(op)
    GalileanBoostOperator(
      velocity=CartesianVel3D( ... )
    )

    An operator with no effect can be simplified:

    >>> op = cxo.GalileanBoostOperator.from_([0, 0, 0], "m/s")
    >>> cxo.simplify_op(op)
    IdentityOperator()

    """
    # Check if the velocity is zero.
    if jnp.allclose(convert(op.velocity, u.Quantity).value, jnp.zeros((3,)), **kwargs):
        return IdentityOperator()
    return op



"""Intra-ecosystem Compatibility."""

__all__: list[str] = []


from typing import TypeAlias

from jaxtyping import Shaped
from plum import convert

import unxt as u

from .cartesian import CartesianPos1D
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.typing import TimeBatchOrScalar

Q1: TypeAlias = Shaped[u.Quantity["length"], "*#batch 1"]


@AbstractOperator.__call__.dispatch
def call(self: AbstractOperator, x: Q1, /) -> Q1:
    """Dispatch to the operator's `__call__` method."""
    return convert(self(CartesianPos1D.from_(x)), u.Quantity)


@AbstractOperator.__call__.dispatch
def call(
    self: AbstractOperator, x: Q1, t: TimeBatchOrScalar, /
) -> tuple[Q1, TimeBatchOrScalar]:
    """Dispatch to the operator's `__call__` method."""
    vec, t = self(CartesianPos1D.from_(x), t)
    return convert(vec, u.Quantity), t
